# pict-rs 0.5.17

pict-rs is a simple image hosting microservice, designed to handle storing and retrieving images,
animations, and videos, as well as providing basic image processing functionality.

## Overview

pict-rs 0.5.17 includes many changes and fixes. I recommend upgrading, especially if you use or plan
to use postgres as your repo.


## Upgrade Notes

The object storage backend has been rewritten since 0.5.16, and existing configurations are not
guaranteed to work. I have tried my best to make the new system work the same way as the old one. If
upgrading causes problems with your object storage and you use vhost style object storage, please
let me know.

For postgres users, a script will run on first launch to attempt to fix the
[utf-8 errors in repo migrations](#utf-8-errors-in-repo-migrations).

In the vast majority of cases, nothing special needs to be done and upgrading is as simple as
pulling and running the new version of pict-rs.


### Major Changes

- [Object Storage Rewrite](#object-storage-rewrite)
- [UTF-8 Errors in Repo Migrations](#utf-8-errors-in-repo-migrations)


### Fixes

- [Incorrect Content-Length HEAD responses](#fix-for-incorrect-content-length-hread-responses)
- [Read filters from the environment](#fix-reading-filters-from-the-environment)
- [More Process Timeouts](#more-process-timeouts)


### Changes

- [Missing Variant Cleanup](#missing-variant-cleanup)
- [Postgres Error Reduction](#postgres-error-reduction)
- [Stream Performance Improvements](#stream-performance-improvements)
- [Pure-Rust Blurhashing](#pure-rust-blurhashing)
- [Include Content-Length in media responses](#include-content-length-in-media-responses)
- [Add integration tests](#add-integration-tests)
- [Add Concurrency to Repo Migration](#add-concurrency-to-repo-migration)
- [Include Service Version in Tracing Context](#include-service-version-in-tracing-context)

### Features

- [Include Blurhash in Details](#include-blurhash-in-details)


## Contributors

- @Nothing4You has a contribution this release: https://git.asonix.dog/asonix/pict-rs/pulls/82


## Details

### Object Storage Rewrite

In order to improve compatibility with existing implementations, and reduce maintenance burden, most
of the object-storage code has been removed in favor of using an off-the-shelf library. The
underlying technologies are largely the same as the previous implementation, so the impact on
pict-rs' dependency tree is minimal.

That said, object_store enforces the use of rustls-native-tls with reqwest, which could potentially
cause issues in some environments. If this change breaks anyone I will work with upstream to resolve
this.


### UTF-8 Errors in Repo Migrations

As it turns out, sometimes UUIDs are made entirely of bytes that are valid utf-8. This causes
problems if a repo migration takes place.

Here is an example of the problem

1. A UUID is generated which is valid utf8 for an Alias and stored in sled.
2. A migration is triggered, and this Alias is read out of sled as a Name, rather than a UUID.
3. This alias is inserted into postgres with a representation that isn't formatted like a UUID.
4. That alias is requested by a user, and doesn't match because the representation is different.

Another example of the problem is the following

1. A Delete Token is generated which is valid utf8 and contains a null byte. Rust says this is fine.
   This delete token is stored in sled.
2. A migration is triggered, and this Delete Token is read out of sled as a Name, rather than a
   UUID.
3. This Delete Token is not inserted into postgres, because postgres does not permit null bytes in
   utf8 strings.

Of these scenarios, only the first can be fixed retroactively. pict-rs now includes a script which
runs on first launch to find and fix any aliases or tokens that have improper representations in the
postgres database.

Any aliases lost to null bytes in past migrations cannot be recovered without running the migration
a second time, but it's likely that a given pict-rs install has diverged enough from it's previous
database that a new migration would cause more problems than it would solve.

In order to fix this going forward, pict-rs always represents aliases and delete tokens that are 16
bytes long as UUIDs in memory, which matches the expected format in the repos. This means any future
migrations will not experience this problem.


### Incorrect Content-Length HEAD responses

Thanks to @Nothing4You for this fix.

HEAD requests are supposed to return the exact headers that a GET request to the same endpoint would
return, but pict-rs was incorrectly returning a `Content-Length: 0` header for these requests. 


### Read filters from the environment

While pict-rs can read most configuration values from the commandline, a configuration file, or
environment variables, the `media.filters` configuration wasn't able to be read from environment
variables. This is fixed now in 0.5.17. This variable now accepts a comma-separated list of
permitted filters for pict-rs.

Example:
```bash
PICTRS__MEDIA__FILTERS=blur,crop
```


### More Process Timeouts

pict-rs now implements more timeouts when dealing with different spawned processes, such as ffmpeg,
imagemagick, and exiftool. This should help terminate and reap misbehaving processes that might
otherwise cause slowdowns.


### Missing Variant Cleanup

This release adds handling for rare cases where variant records exist and refer to missing
files. These records should now be properly cleaned upon request allowing future requests to
re-generate the affected variants.


### Postgres Error Reduction

pict-rs now explicitly tells postgres what to do when inserting records that might cause conflicts
such as duplicate files being uploaded. This prevents postgres from logging errors every time it
happens, but no user-facing functionality should be noticibly changed.


### Stream Performance Improvements

pict-rs has been improved to better drive streams that write into files or standard-in. This means
processing images should be slightly faster.


### Pure-Rust Blurhashing

pict-rs will now avoid calling out to imagemagick for some file types when computing blurhashes,
opting for the `image` library's image decoders instead. It was recently demonstrated that pure-rust
image decoders are now
[on-par or better than equivalent C implementations (reddit)](https://www.redditmedia.com/r/rust/comments/1ha7uyi/memorysafe_png_decoders_now_vastly_outperform_c/),
so where possible, we opt to do the blurhashing entirely in rust. This also improves performance by
avoiding inter-process communication with imagemagick via standard-io.


### Include Content-Length in media responses

Thanks to @Nothing4You for this change.

Previously, pict-rs served files with the `Transfer-Encoding: Chunked` header, which doesn't give an
indication to clients or CDNs how long the served files are. Now, pict-rs will always serve a file
with a `Content-Length` header, improving the ability for CDNs to cache media.


### Add integration tests

pict-rs now has additional tests in the test suite that depend on imagemagick, exiftool, and ffmpeg
being installed on the system running the tests. By default, these tests do not run when
`cargo test` is invoked. In order to run these tests, the `system_deps` configuration flag must be
set.

Example:
```bash
RUSTFLAGS='--cfg system_deps' cargo test
```


### Add Concurrency to Repo Migrations

pict-rs can now migrate multiple hashes between repos concurrently. This doesn't use more threads,
but it does allow pict-rs to get more done on the migration thread. Anecdotally it decreased a
simple 3000 file migration from not-too-fast to pretty-fast.

The default migration concurrency is `32` and that makes a noticable difference. You can change this
with the `-c` flag on the commandline.

```bash
pict-rs \
    migrate-repo -c 128 \
    sled -p data/sled-repo-local \
    postgres -u postgres://pictrs:pictrs@localhost:5432/pictrs
```


### Include Service Version in Tracing Context

pict-rs now reports its version to opentelemetry collectors if those are enabled. This can help keep
track of how timings and spans change after upgrades.


### Include Blurhash in Details

Details responses now include the blurhash of uploaded media. This does cause blurhashes to be
computed far more frequently than before (especially for workloads that didn't use blurhashes), but
in combination with other blurhash improvements, the impact should be small.

Example:
```json
{
  "blurhash": "KMHVC{ShIp.Aoyx@9YSdNY",
  "width": 500,
  "height": 500,
  "frames": 5,
  "content_type": "image/webp",
  "created_at": "2025-04-04T19:08:32.740655791Z",
  "format": "Animation"
}
```
