mod flee;

use std::{
    borrow::Cow,
    fmt::{Debug, Display},
    marker::PhantomData,
    ops::Deref,
};

#[derive(Clone, Debug, Default, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub(crate) struct PathGen;

#[derive(Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub(crate) struct Prefix<Inner, Segment> {
    inner: Inner,
    segment: PhantomData<Segment>,
}

#[derive(Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub(crate) struct Node<Inner, Segment> {
    inner: Inner,
    segment: Segment,
}

impl<'a, Inner, Segment> Debug for Prefix<Inner, Segment>
where
    Inner: Sealed + Debug,
    Segment: PathNode<'a>,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{:?}.{:?}",
            self.inner,
            String::from_utf8_lossy(Segment::NAME)
        )
    }
}

impl<'a, Segment> Debug for Prefix<PathGen, Segment>
where
    Segment: PathNode<'a>,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}", String::from_utf8_lossy(Segment::NAME))
    }
}

impl<'a, Inner, Segment> Debug for Node<Inner, Segment>
where
    Inner: Sealed + Debug,
    Segment: PathNode<'a>,
    Segment: Debug,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{:?}.{:?}.{:?}",
            self.inner,
            String::from_utf8_lossy(Segment::NAME),
            self.segment
        )
    }
}

impl<'a, Segment> Debug for Node<PathGen, Segment>
where
    Segment: PathNode<'a>,
    Segment: Debug,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{:?}.{:?}",
            String::from_utf8_lossy(Segment::NAME),
            self.segment
        )
    }
}

#[derive(Debug)]
pub(crate) struct DeconstructError;

impl Display for DeconstructError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Invalid path format")
    }
}

impl std::error::Error for DeconstructError {}

pub(crate) struct PathParser<Inner> {
    phantom: PhantomData<Inner>,
}

pub(crate) trait PathItem<'a> {
    type Error: std::error::Error;

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a;

    fn to_bytes(&self) -> Cow<'_, [u8]>;
}

pub(crate) trait PathNode<'a>: PathItem<'a> {
    const NAME: &'static [u8];
}

impl PathGen {
    pub(crate) const fn parser() -> PathParser<Self> {
        PathParser {
            phantom: PhantomData,
        }
    }

    pub(crate) const fn push<'a, N>(self, node: N) -> Node<Self, N>
    where
        N: PathNode<'a>,
    {
        Node {
            inner: self,
            segment: node,
        }
    }

    pub(crate) const fn prefix<'a, N>(self) -> Prefix<Self, N>
    where
        N: PathNode<'a>,
    {
        Prefix {
            inner: self,
            segment: PhantomData,
        }
    }
}

impl<Inner, Segment> Node<Inner, Segment> {
    pub(crate) const fn push<'a, N>(self, node: N) -> Node<Self, N>
    where
        N: PathNode<'a>,
    {
        Node {
            inner: self,
            segment: node,
        }
    }

    pub(crate) fn split(self) -> (Inner, Segment) {
        (self.inner, self.segment)
    }

    pub(crate) const fn segment(&self) -> &Segment {
        &self.segment
    }

    pub(crate) const fn prefix<'a, N>(self) -> Prefix<Self, N>
    where
        N: PathNode<'a>,
    {
        Prefix {
            inner: self,
            segment: PhantomData,
        }
    }
}

impl<Inner, Segment> Prefix<Inner, Segment> {
    pub(crate) fn pop(self) -> Inner {
        self.inner
    }
}

impl<Inner, Segment> Prefix<Inner, Segment>
where
    Self: Construct,
{
    pub(crate) fn to_bytes(&self) -> Vec<u8> {
        Construct::construct(self)
    }
}

impl<Inner> PathParser<Inner> {
    pub(crate) const fn push<'a, N>(self) -> PathParser<Node<Inner, N>>
    where
        N: PathNode<'a>,
    {
        PathParser {
            phantom: PhantomData,
        }
    }

    pub(crate) const fn prefix<'a, N>(self) -> PathParser<Prefix<Inner, N>>
    where
        N: PathNode<'a>,
    {
        PathParser {
            phantom: PhantomData,
        }
    }
}

impl<Inner> PathParser<Inner> {
    pub(crate) fn parse<'a, E>(&self, s: &'a [u8]) -> Result<Inner, E>
    where
        Inner: Deconstruct<'a, E> + 'a,
    {
        Inner::deconstruct(s)
    }
}

impl<Inner, Segment> Deref for Node<Inner, Segment> {
    type Target = Segment;

    fn deref(&self) -> &Self::Target {
        self.segment()
    }
}

mod hidden {
    pub(crate) trait Construct {
        fn construct(&self) -> Vec<u8>;
    }

    pub(crate) trait Sealed {}

    pub(crate) trait Deconstruct<'a, E> {
        fn deconstruct(s: &'a [u8]) -> Result<Self, E>
        where
            Self: Sized + 'a;
    }
}
use hidden::{Construct, Deconstruct, Sealed};

impl<'a, Inner, N> Sealed for Node<Inner, N>
where
    Inner: Sealed,
    N: PathNode<'a>,
{
}

impl<'a, N> Sealed for Node<PathGen, N> where N: PathNode<'a> {}

impl<'a, E, Inner, N> Deconstruct<'a, E> for Node<Inner, N>
where
    Inner: Deconstruct<'a, E> + Sealed,
    N: PathNode<'a>,
    E: From<<N as PathItem<'a>>::Error>,
    E: From<DeconstructError>,
{
    fn deconstruct(s: &'a [u8]) -> Result<Self, E>
    where
        Self: Sized + 'a,
    {
        if let Some((rest, value)) = rsplit_once_escaped(s) {
            if let Some((rest, name)) = rsplit_once_escaped(rest) {
                if unescape(name) == N::NAME {
                    let inner = Inner::deconstruct(rest)?;
                    let unescaped = unescape(value);
                    let segment = N::parse(unescaped)?;
                    return Ok(Node { inner, segment });
                }
            }
        }

        Err(DeconstructError.into())
    }
}

impl<'a, E, N> Deconstruct<'a, E> for Node<PathGen, N>
where
    N: PathNode<'a>,
    E: From<<N as PathItem<'a>>::Error>,
    E: From<DeconstructError>,
{
    fn deconstruct(s: &'a [u8]) -> Result<Self, E>
    where
        Self: Sized + 'a,
    {
        if let Some((name, value)) = rsplit_once_escaped(s) {
            if unescape(name) == N::NAME {
                let unescaped = unescape(value);
                let segment = N::parse(unescaped)?;
                return Ok(Node {
                    inner: PathGen,
                    segment,
                });
            }
        }

        Err(DeconstructError.into())
    }
}

impl<'a, Inner, N> Sealed for Prefix<Inner, N>
where
    Inner: Sealed,
    N: PathNode<'a>,
{
}
impl<'a, N> Sealed for Prefix<PathGen, N> where N: PathNode<'a> {}

impl<'a, E, Inner, N> Deconstruct<'a, E> for Prefix<Inner, N>
where
    Inner: Deconstruct<'a, E> + Sealed,
    N: PathNode<'a>,
    E: From<<N as PathItem<'a>>::Error>,
    E: From<DeconstructError>,
{
    fn deconstruct(s: &'a [u8]) -> Result<Self, E>
    where
        Self: Sized + 'a,
    {
        if let Some((rest, name)) = rsplit_once_escaped(s) {
            if unescape(name) == N::NAME {
                let inner = Inner::deconstruct(rest)?;
                return Ok(Prefix {
                    inner,
                    segment: PhantomData,
                });
            }
        }

        Err(DeconstructError.into())
    }
}

impl<'a, E, N> Deconstruct<'a, E> for Prefix<PathGen, N>
where
    N: PathNode<'a>,
    E: From<<N as PathItem<'a>>::Error>,
    E: From<DeconstructError>,
{
    fn deconstruct(s: &'a [u8]) -> Result<Self, E>
    where
        Self: Sized + 'a,
    {
        if unescape(s) == N::NAME {
            return Ok(Prefix {
                inner: PathGen,
                segment: PhantomData,
            });
        }

        Err(DeconstructError.into())
    }
}

impl<'a, Inner, N> Construct for Node<Inner, N>
where
    Inner: Construct,
    N: PathNode<'a>,
{
    fn construct(&self) -> Vec<u8> {
        let mut vec = self.inner.construct();
        vec.extend_from_slice(b".");
        vec.extend_from_slice(&escape(N::NAME));
        vec.extend_from_slice(b".");
        vec.extend_from_slice(&escape(&self.segment.to_bytes()));
        vec
    }
}

impl<'a, N> Construct for Node<PathGen, N>
where
    N: PathNode<'a>,
{
    fn construct(&self) -> Vec<u8> {
        let mut vec = escape(N::NAME).into_owned();
        vec.extend_from_slice(b".");
        vec.extend_from_slice(&escape(&self.segment.to_bytes()));
        vec
    }
}

impl<'a, Inner, N> Construct for Prefix<Inner, N>
where
    Inner: Construct,
    N: PathNode<'a>,
{
    fn construct(&self) -> Vec<u8> {
        let mut vec = self.inner.construct();
        vec.extend_from_slice(b".");
        vec.extend_from_slice(&escape(N::NAME));
        vec
    }
}

impl<'a, N> Construct for Prefix<PathGen, N>
where
    N: PathNode<'a>,
{
    fn construct(&self) -> Vec<u8> {
        escape(N::NAME).into_owned()
    }
}

const DELIMITER: u8 = b'.';
const ESCAPE_TOKEN: u8 = b'\\';

fn escape(s: &[u8]) -> Cow<'_, [u8]> {
    flee::escape(s, DELIMITER, ESCAPE_TOKEN)
}

fn unescape(s: &[u8]) -> Cow<'_, [u8]> {
    flee::unescape(s, DELIMITER, ESCAPE_TOKEN)
}

fn rsplit_once_escaped(s: &[u8]) -> Option<(&[u8], &[u8])> {
    flee::rsplit_once_escaped(s, DELIMITER, ESCAPE_TOKEN)
}

#[cfg(test)]
mod tests {
    use std::{
        borrow::Cow,
        fmt::{Debug, Display},
        num::ParseIntError,
        str::{from_utf8, Utf8Error},
    };

    fn test_pathnode<'a, 'b, N, N2, E>(path_node: N)
    where
        N: PathNode<'a, Error = E> + Debug,
        N2: PathNode<'b, Error = E> + Debug + PartialEq<N> + 'b,
        E: Debug,
    {
        test_roundtrip::<'a, 'b, N, N2, E>(path_node);
    }

    fn test_roundtrip<'a, 'b, T, T2, E>(item: T)
    where
        T: PathItem<'a, Error = E> + Debug,
        T2: PathItem<'b, Error = E> + Debug + PartialEq<T> + 'b,
        E: Debug,
    {
        let vec = item.to_bytes().to_vec();
        let round_trip = T2::parse(Cow::Owned(vec.clone())).unwrap();
        let plus_one = round_trip.to_bytes().to_vec();

        assert_eq!(round_trip, item);
        assert_eq!(vec, plus_one);
    }

    use super::{PathGen, PathItem, PathNode};

    #[test]
    fn construct_prefix() {
        let s = PathGen
            .push(Deck("deck-id".into()))
            .prefix::<Key>()
            .to_bytes();

        assert_eq!(&s, b"deck.deck-id.key")
    }

    #[test]
    fn construct_escaped_prefix() {
        let s = PathGen
            .push(Deck("deck.id".into()))
            .prefix::<Key>()
            .to_bytes();

        assert_eq!(&s, b"deck.deck\\.id.key")
    }

    #[test]
    fn deconstruct_prefix() {
        let parser = PathGen::parser().push::<Deck>().prefix::<Key>();

        let s = b"deck.deck-id.key";

        let path = parser.parse::<Box<dyn std::error::Error>>(s).unwrap();

        assert_eq!(path, PathGen.push(Deck("deck-id".into())).prefix::<Key>());
    }

    #[test]
    fn deconstruct_escaped_prefix() {
        let parser = PathGen::parser().push::<Deck>().prefix::<Key>();

        let s = b"deck.deck\\.id.key";

        let path = parser.parse::<Box<dyn std::error::Error>>(s).unwrap();

        assert_eq!(path, PathGen.push(Deck("deck.id".into())).prefix::<Key>());
    }

    #[test]
    fn test_parts() {
        let string = String::from("hello");
        let brw: &str = &string;

        test_pathnode::<_, Deck<'_>, _>(Deck(Cow::Borrowed(brw)));
        test_pathnode::<_, Key, _>(Key(5));
    }

    #[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
    struct Deck<'a>(Cow<'a, str>);
    #[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
    struct Key(u8);
    #[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
    struct Command;
    #[derive(Debug)]
    struct CommandError;

    impl Display for CommandError {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            write!(f, "string did not match 'command'")
        }
    }

    impl std::error::Error for CommandError {}

    impl<'a> PathItem<'a> for Deck<'a> {
        type Error = Utf8Error;

        fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
        where
            Self: Sized + 'a,
        {
            let cow = match bytes {
                Cow::Owned(bytes) => {
                    from_utf8(&bytes)?;
                    Cow::Owned(String::from_utf8(bytes).unwrap())
                }
                Cow::Borrowed(bytes) => Cow::Borrowed(from_utf8(bytes)?),
            };

            Ok(Deck(cow))
        }

        fn to_bytes(&self) -> Cow<'_, [u8]> {
            Cow::Borrowed(self.0.as_bytes())
        }
    }

    impl<'a> PathItem<'a> for Key {
        type Error = ParseIntError;

        fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
        where
            Self: Sized + 'a,
        {
            Ok(Key(String::from_utf8_lossy(&bytes).parse()?))
        }

        fn to_bytes(&self) -> Cow<'_, [u8]> {
            Cow::Owned(self.0.to_string().into_bytes())
        }
    }

    impl<'a> PathItem<'a> for Command {
        type Error = CommandError;

        fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
        where
            Self: Sized + 'a,
        {
            if bytes.as_ref() == b"cmd" {
                return Ok(Command);
            }

            Err(CommandError)
        }

        fn to_bytes(&self) -> Cow<'_, [u8]> {
            Cow::Borrowed(b"cmd")
        }
    }

    impl<'a> PathNode<'a> for Deck<'a> {
        const NAME: &'static [u8] = b"deck";
    }

    impl<'a> PathNode<'a> for Key {
        const NAME: &'static [u8] = b"key";
    }

    impl<'a> PathNode<'a> for Command {
        const NAME: &'static [u8] = b"command";
    }
}
