use crate::{
    details::DetailsInner,
    error_code::{ErrorCode, OwnedErrorCode},
    future::WithPollTimer,
    path_gen::{DeconstructError, PathItem},
    serde_str::Serde,
    stream::from_unsend_iterator,
};
use dashmap::DashMap;
use fjall::{
    Config, PersistMode, TransactionalKeyspace, TransactionalPartitionHandle, WriteTransaction,
};
use futures_core::stream::LocalBoxStream;
use std::{
    array::TryFromSliceError,
    io::Write,
    str::Utf8Error,
    sync::{
        atomic::{AtomicU64, Ordering},
        Arc,
    },
    time::Duration,
};
use url::Url;
use uuid::Uuid;

use super::{
    metrics::{PopMetricsGuard, PushMetricsGuard, WaitMetricsGuard},
    notification_map::NotificationMap,
    Alias, AliasAccessRepo, AliasAlreadyExists, AliasRepo, BaseRepo, DeleteToken, Details,
    DetailsRepo, FullRepo, Hash, HashAlreadyExists, HashPage, HashRepo, JobId, JobResult,
    NotificationEntry, OrderedHash, ProxyAlreadyExists, ProxyRepo, QueueRepo, RepoError,
    SettingsRepo, StoreMigrationRepo, UploadId, UploadRepo, UploadResult, VariantAccessRepo,
    VariantAlreadyExists, VariantRepo,
};

mod alias;
mod bidirectional_mapping;
mod blurhash;
mod delete_token;
mod details;
mod empty;
mod hash;
mod hash_alias;
mod hash_variant;
mod healthz;
mod identifier;
mod job;
mod job_id;
mod job_state;
mod mappings;
mod proxy;
mod queue;
mod queue_job_key;
mod retry_count;
mod settings;
mod timestamp;
mod tree;
mod unique_job;
mod upload;
mod upload_value;
mod variant;

use bidirectional_mapping::{Bidirectional, MappingAlreadyExists};
use blurhash::Blurhash;
use empty::Empty;
use hash_alias::{HashAlias, HashAliasError};
use hash_variant::{HashVariant, HashVariantError};
use healthz::{Healthz, HealthzKey};
use identifier::{Identifier, MigratedIdentifier, MotionIdentifier};
use job::Job;
use job_state::JobState;
use mappings::*;
use proxy::{Proxy, ProxyError};
use queue::Queue;
use queue_job_key::{QueueJobKey, QueueJobKeyError};
use retry_count::RetryCount;
use settings::{Setting, SettingsValue};
use timestamp::{Accessed, Created, TimeError, Timestamp};
use tree::Mapping;
use unique_job::UniqueJob;
use upload_value::UploadValue;
use variant::Variant;

#[derive(Clone)]
pub(crate) struct FjallRepo {
    inner: Arc<FjallInner>,
}

#[derive(Clone)]
struct FjallInner {
    partition: TransactionalPartitionHandle,
    upload_notifications: NotificationMap,
    job_notifications: NotificationMap,
    variant_notifications: NotificationMap,
    variant_process_map: DashMap<(Hash, String), time::OffsetDateTime>,
    healthz_count: Arc<AtomicU64>,
    db: TransactionalKeyspace,
}

impl FjallInner {
    fn write_tx(&self) -> WriteTransaction<'_> {
        self.db.write_tx().durability(Some(PersistMode::SyncAll))
    }
}

#[derive(Debug, thiserror::Error)]
pub(crate) enum FjallError {
    #[error("Error in fjall")]
    Fjall(#[from] fjall::Error),

    #[error("Invalid key path")]
    Deconstruct(#[from] DeconstructError),

    #[error("Invalid alias")]
    Alias,

    #[error("Error parsing blurhash")]
    Blurhash(#[source] Utf8Error),

    #[error("Invalid delete token")]
    DeleteToken,

    #[error("Error parsing details")]
    Details(#[source] serde_json::Error),

    #[error("Buffer is not empty")]
    Empty,

    #[error("Invalid hash")]
    Hash,

    #[error("Invalid HashAlias")]
    HashAlias(#[from] HashAliasError),

    #[error("Invalid HashVariant")]
    HashVariant(#[from] HashVariantError),

    #[error("Invalid healthz")]
    Healthz(#[source] TryFromSliceError),

    #[error("Invalid healthz key")]
    HealthzKey,

    #[error("Invalid identifier")]
    Identifier(#[source] Utf8Error),

    #[error("Invalid job")]
    Job(#[source] serde_json::Error),

    #[error("Invalid job id")]
    JobId(#[source] TryFromSliceError),

    #[error("Invalid job state")]
    JobState(#[source] TryFromSliceError),

    #[error("Invalid proxy")]
    Proxy(#[from] ProxyError),

    #[error("Invalid Queue")]
    Queue(#[source] Utf8Error),

    #[error("Invalid Queue Job key")]
    QueueJob(#[from] QueueJobKeyError),

    #[error("Invalid retry count")]
    RetryCount(#[source] TryFromSliceError),

    #[error("Invalid setting")]
    Setting(#[source] Utf8Error),

    #[error("Invalid timestamp")]
    Timestamp(#[from] TimeError),

    #[error("Invalid unique job key")]
    UniqueJob(#[source] Utf8Error),

    #[error("Invalid upload ID")]
    UploadId(#[source] uuid::Error),

    #[error("Invalid Upload")]
    Upload(#[source] serde_json::Error),

    #[error("Invalid variant")]
    Variant(#[source] Utf8Error),
}

impl From<std::convert::Infallible> for FjallError {
    fn from(value: std::convert::Infallible) -> Self {
        match value {}
    }
}

impl FjallError {
    pub(super) const fn error_code(&self) -> ErrorCode {
        match self {
            Self::Fjall(_)
            | Self::Deconstruct(_)
            | Self::Alias
            | Self::Blurhash(_)
            | Self::DeleteToken
            | Self::Empty
            | Self::Hash
            | Self::HashAlias(_)
            | Self::HashVariant(_)
            | Self::Healthz(_)
            | Self::HealthzKey
            | Self::Identifier(_)
            | Self::JobId(_)
            | Self::JobState(_)
            | Self::Proxy(_)
            | Self::Queue(_)
            | Self::QueueJob(_)
            | Self::RetryCount(_)
            | Self::Setting(_)
            | Self::Timestamp(_)
            | Self::UniqueJob(_)
            | Self::UploadId(_)
            | Self::Variant(_) => ErrorCode::FJALL_ERROR,
            Self::Details(_) => ErrorCode::EXTRACT_DETAILS,
            Self::Job(_) => ErrorCode::EXTRACT_JOB,
            Self::Upload(_) => ErrorCode::EXTRACT_UPLOAD_RESULT,
        }
    }
}

type HashBounds = (std::ops::Bound<Created>, std::ops::Bound<Created>);

impl FjallRepo {
    #[tracing::instrument]
    pub(crate) fn build(
        crate::config::Fjall {
            mut fjall_path,
            debug_dump_db_path,
        }: crate::config::Fjall,
    ) -> color_eyre::Result<Self> {
        fjall_path.push("v0.5.18");

        let db = Config::new(fjall_path).open_transactional()?;

        if db.partition_count() == 0 {
            tracing::info!("No existing Fjall database found, creating an empty Fjall repo");
        }

        let partition = db.open_partition("fjall-partition", Default::default())?;

        let tx = db.read_tx();

        if let Some(dump_path) = debug_dump_db_path {
            let mut file = std::fs::File::options()
                .create(true)
                .write(true)
                .truncate(true)
                .open(dump_path)?;

            let mut writer = std::io::BufWriter::new(&mut file);
            for res in tx.iter(&partition) {
                let (key, value) = res?;

                writeln!(&mut writer, "{:?}", mappings::RowDebugger(&key, &value))?;
            }

            writer.flush()?;
            drop(writer);
            file.flush()?;
        }

        Ok(Self {
            inner: Arc::new(FjallInner {
                partition,
                upload_notifications: NotificationMap::new(),
                job_notifications: NotificationMap::new(),
                variant_notifications: NotificationMap::new(),
                variant_process_map: DashMap::new(),
                healthz_count: Arc::new(AtomicU64::new(0)),
                db,
            }),
        })
    }

    async fn op<F, Out>(&self, function: F) -> Result<Out, RepoError>
    where
        F: FnOnce(&FjallInner) -> Result<Out, FjallError> + Send + 'static,
        Out: Send + 'static,
    {
        let inner = Arc::clone(&self.inner);

        crate::sync::spawn_blocking("fjall-io", move || (function)(&inner))
            .await
            .map_err(|_| RepoError::Canceled)?
            .map_err(RepoError::FjallError)
    }

    async fn bounded_hashes(
        &self,
        page_range: HashBounds,
        prev_range: Option<HashBounds>,
        limit: usize,
    ) -> Result<HashPage, RepoError> {
        self.op(move |inner| {
            let created_mapping =
                Bidirectional::new(&inner.partition, HashCreatedMapping, CreatedHashMapping);

            let tx = inner.db.read_tx();

            let page_iter = created_mapping.range_by_value(&tx, page_range);

            let page_iter = page_iter
                .rev()
                .map(|res| {
                    res.map(|(hash, Created(Timestamp(timestamp, ..)))| OrderedHash {
                        hash,
                        timestamp,
                    })
                })
                .take(limit + 1);

            let prev = prev_range
                .and_then(|prev_range| {
                    created_mapping
                        .range_by_value(&tx, prev_range)
                        .map(|res| {
                            res.map(|(hash, Created(Timestamp(timestamp, ..)))| OrderedHash {
                                hash,
                                timestamp,
                            })
                        })
                        .take(limit + 1)
                        .last()
                })
                .transpose()?;

            let mut hashes = page_iter.collect::<Result<Vec<_>, _>>()?;

            let next = if hashes.len() > limit {
                hashes.pop()
            } else {
                None
            };

            let prev = if prev.as_ref() == hashes.first() {
                None
            } else {
                prev
            };

            Ok(HashPage {
                limit,
                prev: prev.map(|OrderedHash { hash, .. }| hash),
                next: next.map(|OrderedHash { hash, .. }| hash),
                hashes: hashes
                    .into_iter()
                    .map(|OrderedHash { hash, .. }| hash)
                    .collect(),
            })
        })
        .await
    }
}

impl BaseRepo for FjallRepo {}

#[async_trait::async_trait(?Send)]
impl FullRepo for FjallRepo {
    async fn health_check(&self) -> Result<(), RepoError> {
        self.op(|inner| {
            let next = inner.healthz_count.fetch_add(1, Ordering::Relaxed);
            let mapping = HealthzMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            mapping.insert(&mut tx, HealthzKey, Healthz(next));

            tx.commit()?;

            Ok(())
        })
        .await
    }
}

#[async_trait::async_trait(?Send)]
impl ProxyRepo for FjallRepo {
    async fn relate_url(
        &self,
        url: Url,
        alias: Alias,
    ) -> Result<Result<(), ProxyAlreadyExists>, RepoError> {
        self.op(move |inner| {
            let mut tx = inner.write_tx();

            let mapping =
                Bidirectional::new(&inner.partition, ProxyAliasMapping, AliasProxyMapping);

            match mapping.insert_unique(&mut tx, Proxy(url), alias)? {
                Ok(()) => {
                    tx.commit()?;
                    Ok(Ok(()))
                }
                Err(MappingAlreadyExists) => {
                    tx.rollback();
                    Ok(Err(ProxyAlreadyExists))
                }
            }
        })
        .await
    }

    async fn related(&self, url: Url) -> Result<Option<Alias>, RepoError> {
        self.op(move |inner| {
            let tx = inner.db.read_tx();

            let mapping =
                Bidirectional::new(&inner.partition, ProxyAliasMapping, AliasProxyMapping);

            let alias = mapping.get(&tx, Proxy(url))?;

            Ok(alias)
        })
        .await
    }

    async fn remove_relation(&self, alias: Alias) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mut tx = inner.write_tx();

            let mapping =
                Bidirectional::new(&inner.partition, ProxyAliasMapping, AliasProxyMapping);

            mapping.remove_by_value(&mut tx, alias)?;

            Ok(())
        })
        .await
    }
}

#[async_trait::async_trait(?Send)]
impl AliasAccessRepo for FjallRepo {
    #[tracing::instrument(level = "debug", skip(self))]
    async fn set_accessed_alias(
        &self,
        alias: Alias,
        accessed: time::OffsetDateTime,
    ) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping =
                Bidirectional::new(&inner.partition, AliasAccessMapping, AccessAliasMapping);
            let value = Accessed(Timestamp(accessed, alias.to_bytes()));

            let mut tx = inner.write_tx();

            mapping.overwrite(&mut tx, alias, value)?;

            Ok(())
        })
        .await
    }

    async fn alias_accessed_at(
        &self,
        alias: Alias,
    ) -> Result<Option<time::OffsetDateTime>, RepoError> {
        self.op(move |inner| {
            let mapping =
                Bidirectional::new(&inner.partition, AliasAccessMapping, AccessAliasMapping);

            let tx = inner.db.read_tx();

            let opt = mapping.get(&tx, alias)?;

            Ok(opt.map(|Accessed(Timestamp(accessed, ..))| accessed))
        })
        .await
    }

    #[tracing::instrument(level = "debug", skip(self))]
    async fn older_aliases(
        &self,
        timestamp: time::OffsetDateTime,
    ) -> Result<LocalBoxStream<'static, Result<Alias, RepoError>>, RepoError> {
        use std::ops::Bound;

        let inner = Arc::clone(&self.inner);

        Ok(Box::pin(from_unsend_iterator(
            move || {
                let mapping =
                    Bidirectional::new(&inner.partition, AliasAccessMapping, AccessAliasMapping);

                let accessed = Accessed(Timestamp(timestamp, vec![]));

                let tx = inner.db.read_tx();

                mapping
                    .range_by_value(&tx, (Bound::Unbounded, Bound::Included(accessed)))
                    .map(|res| {
                        res.map(|(k, _)| k)
                            .map_err(FjallError::from)
                            .map_err(RepoError::FjallError)
                    })
            },
            8,
        )))
    }

    #[tracing::instrument(level = "debug", skip(self))]
    async fn remove_alias_access(&self, alias: Alias) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping =
                Bidirectional::new(&inner.partition, AliasAccessMapping, AccessAliasMapping);

            let mut tx = inner.write_tx();

            mapping.remove(&mut tx, alias)?;

            tx.commit()?;

            Ok(())
        })
        .await
    }
}

#[async_trait::async_trait(?Send)]
impl VariantAccessRepo for FjallRepo {
    #[tracing::instrument(level = "debug", skip(self))]
    async fn set_accessed_variant(
        &self,
        hash: Hash,
        variant: String,
        accessed: time::OffsetDateTime,
    ) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping =
                Bidirectional::new(&inner.partition, VariantAccessMapping, AccessVariantMapping);

            let key = HashVariant(hash, variant);
            let value = Accessed(Timestamp(accessed, key.to_bytes().to_vec()));

            let mut tx = inner.write_tx();

            mapping.overwrite(&mut tx, key, value)?;

            tx.commit()?;

            Ok(())
        })
        .await
    }

    #[tracing::instrument(level = "debug", skip(self))]
    async fn variant_accessed_at(
        &self,
        hash: Hash,
        variant: String,
    ) -> Result<Option<time::OffsetDateTime>, RepoError> {
        self.op(move |inner| {
            let mapping =
                Bidirectional::new(&inner.partition, VariantAccessMapping, AccessVariantMapping);

            let key = HashVariant(hash, variant);

            let tx = inner.db.read_tx();

            let opt = mapping.get(&tx, key)?;

            Ok(opt.map(|Accessed(Timestamp(accessed, ..))| accessed))
        })
        .await
    }

    #[tracing::instrument(level = "debug", skip(self))]
    async fn older_variants(
        &self,
        timestamp: time::OffsetDateTime,
    ) -> Result<LocalBoxStream<'static, Result<(Hash, String), RepoError>>, RepoError> {
        use std::ops::Bound;

        let inner = Arc::clone(&self.inner);

        Ok(Box::pin(from_unsend_iterator(
            move || {
                let mapping = Bidirectional::new(
                    &inner.partition,
                    VariantAccessMapping,
                    AccessVariantMapping,
                );

                let accessed = Accessed(Timestamp(timestamp, vec![]));

                let tx = inner.db.read_tx();

                mapping
                    .range_by_value(&tx, (Bound::Unbounded, Bound::Included(accessed)))
                    .map(|res| {
                        res.map(|(k, _)| k)
                            .map_err(FjallError::from)
                            .map_err(RepoError::FjallError)
                            .map(|HashVariant(hash, variant)| (hash, variant))
                    })
            },
            8,
        )))
    }

    #[tracing::instrument(level = "debug", skip(self))]
    async fn remove_variant_access(&self, hash: Hash, variant: String) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping =
                Bidirectional::new(&inner.partition, VariantAccessMapping, AccessVariantMapping);

            let key = HashVariant(hash, variant);

            let mut tx = inner.write_tx();

            mapping.remove(&mut tx, key)?;

            tx.commit()?;

            Ok(())
        })
        .await
    }
}

#[derive(Debug, serde::Deserialize, serde::Serialize)]
enum InnerUploadResult {
    Success {
        alias: Serde<Alias>,
        token: Serde<DeleteToken>,
    },
    Failure {
        message: String,
        code: OwnedErrorCode,
    },
}

impl From<UploadResult> for InnerUploadResult {
    fn from(u: UploadResult) -> Self {
        match u {
            UploadResult::Success { alias, token } => InnerUploadResult::Success {
                alias: Serde::new(alias),
                token: Serde::new(token),
            },
            UploadResult::Failure { message, code } => InnerUploadResult::Failure { message, code },
        }
    }
}

impl From<InnerUploadResult> for UploadResult {
    fn from(i: InnerUploadResult) -> Self {
        match i {
            InnerUploadResult::Success { alias, token } => UploadResult::Success {
                alias: Serde::into_inner(alias),
                token: Serde::into_inner(token),
            },
            InnerUploadResult::Failure { message, code } => UploadResult::Failure { message, code },
        }
    }
}

#[async_trait::async_trait(?Send)]
impl UploadRepo for FjallRepo {
    #[tracing::instrument(level = "trace", skip(self))]
    async fn create_upload(&self) -> Result<UploadId, RepoError> {
        self.op(move |inner| {
            let mapping = UploadMapping.tree(&inner.partition);

            let upload_id = UploadId::generate();

            let mut tx = inner.write_tx();

            mapping.insert(&mut tx, upload_id, UploadValue::Pending);

            tx.commit()?;

            Ok(upload_id)
        })
        .await
    }

    #[tracing::instrument(skip(self))]
    async fn wait(&self, upload_id: UploadId) -> Result<UploadResult, RepoError> {
        let guard = WaitMetricsGuard::guard();

        loop {
            tracing::trace!("wait: looping");

            let mut interest = self
                .inner
                .upload_notifications
                .register_interest(Arc::from(upload_id.to_string()));
            let notified = interest.notified_timeout(Duration::from_secs(5));

            let maybe_result = self
                .op(move |inner| {
                    let mapping = UploadMapping.tree(&inner.partition);

                    let tx = inner.db.read_tx();

                    match mapping.get(&tx, upload_id)? {
                        Some(UploadValue::Pending) => Ok(Ok(None)),
                        Some(UploadValue::Result(inner)) => Ok(Ok(Some(inner))),
                        None => Ok(Err(RepoError::AlreadyClaimed)),
                    }
                })
                .await??;

            match maybe_result {
                Some(result) => {
                    guard.disarm();
                    return Ok(result.into());
                }
                None => match notified.await {
                    Ok(()) => tracing::debug!("Notified"),
                    Err(_) => tracing::trace!("Timed out"),
                },
            }
        }
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn claim(&self, upload_id: UploadId) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping = UploadMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            mapping.remove(&mut tx, upload_id);

            tx.commit()?;

            Ok(())
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self, result))]
    async fn complete_upload(
        &self,
        upload_id: UploadId,
        result: UploadResult,
    ) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping = UploadMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            mapping.insert(&mut tx, upload_id, UploadValue::Result(result.into()));

            tx.commit()?;

            inner.upload_notifications.notify(&upload_id.to_string());

            Ok(())
        })
        .await
    }
}

#[async_trait::async_trait(?Send)]
impl QueueRepo for FjallRepo {
    async fn queue_length(&self) -> Result<u64, RepoError> {
        self.op(move |inner| {
            let job_mapping = JobMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let length = job_mapping.prefix(&tx, None).count();

            Ok(length as _)
        })
        .await
    }

    #[tracing::instrument(skip(self))]
    async fn push(
        &self,
        queue_name: &'static str,
        job: serde_json::Value,
        unique_key: Option<&'static str>,
    ) -> Result<Option<JobId>, RepoError> {
        let metrics_guard = PushMetricsGuard::guard(queue_name);

        let job_id = self
            .op(move |inner| {
                let queue_mapping = Bidirectional::new(
                    &inner.partition,
                    UniqueJobQueueMapping,
                    QueueUniqueJobMapping,
                );
                let job_mapping = JobMapping.tree(&inner.partition);
                let state_mapping = JobStateMapping.tree(&inner.partition);

                let id = JobId::gen();

                let mut tx = inner.write_tx();

                let queue_job_key = QueueJobKey::new(queue_name, id);

                if let Some(unique_key) = unique_key {
                    match queue_mapping.insert_unique(
                        &mut tx,
                        UniqueJob::New(unique_key),
                        queue_job_key.clone(),
                    )? {
                        Ok(()) => {}
                        Err(MappingAlreadyExists) => return Ok(None),
                    }
                }

                job_mapping.insert(&mut tx, queue_job_key.clone(), Job(job));
                state_mapping.insert(&mut tx, queue_job_key, JobState::Pending);

                tx.commit()?;

                inner.job_notifications.notify(queue_name);

                Ok(Some(id))
            })
            .await?;

        metrics_guard.disarm();

        Ok(job_id)
    }

    #[tracing::instrument(skip_all, fields(job_id))]
    async fn pop(
        &self,
        queue_name: &'static str,
        worker_id: Uuid,
    ) -> Result<(JobId, serde_json::Value), RepoError> {
        let metrics_guard = PopMetricsGuard::guard(queue_name);

        loop {
            tracing::trace!("pop: looping");

            let mut interest = self
                .inner
                .job_notifications
                .register_interest(Arc::from(queue_name));

            let opt = self
                .op(move |inner| {
                    let job_mapping = JobMapping.tree(&inner.partition);
                    let state_mapping = JobStateMapping.tree(&inner.partition);

                    let now = time::OffsetDateTime::now_utc();

                    let read_tx = inner.db.read_tx();

                    let queue_prefix = QueueJobKey::prefix(queue_name);

                    for res in state_mapping.prefix(&read_tx, Some(queue_prefix)) {
                        let (job_key, previous_state) = res?;

                        if !previous_state.is_pending()
                            && !previous_state.is_older_than(time::Duration::seconds(30), now)
                        {
                            // job is in progress already
                            continue;
                        }

                        let new_state = JobState::running(worker_id);

                        let mut tx = inner.write_tx();

                        let opt = state_mapping.write_get(&mut tx, job_key.clone())?;

                        if opt == Some(previous_state) {
                            state_mapping.insert(&mut tx, job_key.clone(), new_state);
                            tx.commit()?;
                        } else {
                            tx.rollback();
                            tracing::debug!("contested");
                            continue;
                        }

                        let job_id = job_key.1;

                        if let Some(Job(value)) = job_mapping.get(&read_tx, job_key)? {
                            return Ok(Some((job_id, value)));
                        }

                        tracing::debug!("Failed to get job value");
                    }

                    Ok(None)
                })
                .await?;

            if let Some((job_id, job_json)) = opt {
                tracing::Span::current().record("job_id", format!("{}", job_id.0));

                metrics_guard.disarm();
                tracing::debug!("{job_json}");
                return Ok((job_id, job_json));
            }

            match interest
                .notified_timeout(Duration::from_secs(30))
                .with_poll_timer("fjall-pop-notify")
                .await
            {
                Ok(()) => tracing::debug!("Notified"),
                Err(_) => tracing::trace!("Timed out"),
            }
        }
    }

    #[tracing::instrument(skip(self, worker_id))]
    async fn heartbeat(
        &self,
        queue_name: &'static str,
        worker_id: Uuid,
        job_id: JobId,
    ) -> Result<(), RepoError> {
        self.op(move |inner| {
            let state_mapping = JobStateMapping.tree(&inner.partition);

            let job_key = QueueJobKey::new(queue_name, job_id);

            let mut tx = inner.write_tx();

            if state_mapping.contains_key(&mut tx, job_key.clone())? {
                state_mapping.insert(&mut tx, job_key, JobState::running(worker_id));
            }

            tx.commit()?;

            Ok(())
        })
        .await
    }

    #[tracing::instrument(skip(self, _worker_id))]
    async fn complete_job(
        &self,
        queue_name: &'static str,
        _worker_id: Uuid,
        job_id: JobId,
        job_status: JobResult,
    ) -> Result<(), RepoError> {
        let retried = self
            .op(move |inner| {
                let queue_mapping = Bidirectional::new(
                    &inner.partition,
                    UniqueJobQueueMapping,
                    QueueUniqueJobMapping,
                );
                let job_mapping = JobMapping.tree(&inner.partition);
                let state_mapping = JobStateMapping.tree(&inner.partition);
                let retry_mapping = JobRetryMapping.tree(&inner.partition);

                let retry = matches!(job_status, JobResult::Failure);

                let mut tx = inner.write_tx();

                let key = QueueJobKey::new(queue_name, job_id);

                let retry_count = retry_mapping
                    .write_get(&mut tx, key.clone())?
                    .map(|RetryCount(count)| count)
                    .unwrap_or(5)
                    .saturating_sub(1);

                if retry_count > 0 && retry {
                    retry_mapping.insert(&mut tx, key, RetryCount(retry_count));
                    tx.commit()?;

                    Ok(true)
                } else {
                    job_mapping.remove(&mut tx, key.clone());
                    state_mapping.remove(&mut tx, key.clone());
                    retry_mapping.remove(&mut tx, key.clone());
                    queue_mapping.remove_by_value(&mut tx, key)?;

                    tx.commit()?;

                    Ok(false)
                }
            })
            .await?;

        match job_status {
            JobResult::Success => tracing::debug!("completed {job_id:?}"),
            JobResult::Failure if retried => {
                tracing::info!("{job_id:?} failed, marked for retry")
            }
            JobResult::Failure => tracing::warn!("{job_id:?} failed permantently"),
            JobResult::Aborted => tracing::warn!("{job_id:?} dead"),
        }

        Ok(())
    }
}

#[async_trait::async_trait(?Send)]
impl SettingsRepo for FjallRepo {
    #[tracing::instrument(level = "trace", skip(self, value))]
    async fn set(&self, key: &'static str, value: Arc<[u8]>) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping = SettingsMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            mapping.insert(&mut tx, Setting::New(key), SettingsValue(value));

            tx.commit()?;

            Ok(())
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn get(&self, key: &'static str) -> Result<Option<Arc<[u8]>>, RepoError> {
        self.op(move |inner| {
            let mapping = SettingsMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let opt = mapping.get(&tx, Setting::New(key))?;

            Ok(opt.map(|SettingsValue(v)| v))
        })
        .await
    }
}

#[async_trait::async_trait(?Send)]
impl DetailsRepo for FjallRepo {
    #[tracing::instrument(level = "trace", skip(self))]
    async fn relate_details(
        &self,
        identifier: &Arc<str>,
        details: &Details,
    ) -> Result<(), RepoError> {
        let identifier = Identifier(Arc::clone(identifier));
        let details = details.inner.clone();

        self.op(move |inner| {
            let mapping = IdentifierDetailsMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            mapping.insert(&mut tx, identifier, details);

            tx.commit()?;

            Ok(())
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn details(&self, identifier: &Arc<str>) -> Result<Option<Details>, RepoError> {
        let identifier = Identifier(Arc::clone(identifier));

        self.op(move |inner| {
            let mapping = IdentifierDetailsMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let opt = mapping.get(&tx, identifier)?;

            Ok(opt.map(|inner| Details { inner }))
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn cleanup_details(&self, identifier: &Arc<str>) -> Result<(), RepoError> {
        let identifier = Identifier(Arc::clone(identifier));

        self.op(move |inner| {
            let mapping = IdentifierDetailsMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            mapping.remove(&mut tx, identifier);

            tx.commit()?;

            Ok(())
        })
        .await
    }
}

#[async_trait::async_trait(?Send)]
impl StoreMigrationRepo for FjallRepo {
    async fn is_continuing_migration(&self) -> Result<bool, RepoError> {
        self.op(|inner| {
            let mapping = MigratedIdentifierMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let nonempty = mapping
                .prefix(&tx, None)
                .filter_map(Result::ok)
                .any(|_| true);

            Ok(nonempty)
        })
        .await
    }

    async fn mark_migrated(
        &self,
        old_identifier: &Arc<str>,
        new_identifier: &Arc<str>,
    ) -> Result<(), RepoError> {
        let key = MigratedIdentifier::new(new_identifier);
        let value = Identifier::new(old_identifier);

        self.op(move |inner| {
            let mapping = MigratedIdentifierMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            mapping.insert(&mut tx, key, value);

            tx.commit()?;

            Ok(())
        })
        .await
    }

    async fn is_migrated(&self, identifier: &Arc<str>) -> Result<bool, RepoError> {
        let key = MigratedIdentifier::new(identifier);

        self.op(move |inner| {
            let mapping = MigratedIdentifierMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let contains = mapping.read_contains_key(&tx, key)?;

            Ok(contains)
        })
        .await
    }

    async fn clear(&self) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping = MigratedIdentifierMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let mut maybe_write_tx = None;
            let mut count = 0;

            for res in mapping.prefix(&tx, None) {
                let (key, _value) = res?;

                let write_tx = maybe_write_tx.get_or_insert_with(move || inner.write_tx());

                mapping.remove(write_tx, key);

                count += 1;

                if count == 1000 {
                    count = 0;
                    if let Some(write_tx) = maybe_write_tx.take() {
                        write_tx.commit()?;
                    }
                }
            }

            Ok(())
        })
        .await
    }
}

#[async_trait::async_trait(?Send)]
impl HashRepo for FjallRepo {
    async fn size(&self) -> Result<u64, RepoError> {
        self.op(|inner| {
            let created_mapping =
                Bidirectional::new(&inner.partition, HashCreatedMapping, CreatedHashMapping);

            let tx = inner.db.read_tx();

            let len = created_mapping.len(&tx, None)?;

            Ok(len as _)
        })
        .await
    }

    async fn bound(&self, hash: Hash) -> Result<Option<OrderedHash>, RepoError> {
        self.op(move |inner| {
            let created_mapping =
                Bidirectional::new(&inner.partition, HashCreatedMapping, CreatedHashMapping);

            let tx = inner.db.read_tx();

            let opt = created_mapping
                .get(&tx, hash.clone())?
                .map(|Created(Timestamp(timestamp, ..))| OrderedHash { hash, timestamp });

            Ok(opt)
        })
        .await
    }

    async fn hashes_ordered(
        &self,
        bound: Option<OrderedHash>,
        limit: usize,
    ) -> Result<HashPage, RepoError> {
        use std::ops::Bound;

        let (page_range, prev_range) = match &bound {
            Some(ordered_hash) => {
                let timestamp = Timestamp(ordered_hash.timestamp, ordered_hash.hash.to_bytes());
                (
                    (
                        Bound::Unbounded,
                        Bound::Included(Created(timestamp.clone())),
                    ),
                    Some((Bound::Included(Created(timestamp)), Bound::Unbounded)),
                )
            }
            None => ((Bound::Unbounded, Bound::Unbounded), None),
        };

        self.bounded_hashes(page_range, prev_range, limit).await
    }

    async fn hash_page_by_date(
        &self,
        date: time::OffsetDateTime,
        limit: usize,
    ) -> Result<HashPage, RepoError> {
        use std::ops::Bound;

        let timestamp = Timestamp(date, vec![]);

        let page_range = (
            Bound::Unbounded,
            Bound::Included(Created(timestamp.clone())),
        );
        let prev_range = Some((Bound::Included(Created(timestamp)), Bound::Unbounded));

        self.bounded_hashes(page_range, prev_range, limit).await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn create_hash_with_timestamp(
        &self,
        hash: Hash,
        identifier: &Arc<str>,
        timestamp: time::OffsetDateTime,
    ) -> Result<Result<(), HashAlreadyExists>, RepoError> {
        let created_key = Timestamp(timestamp, hash.to_bytes());
        let identifier = Identifier(Arc::clone(identifier));

        self.op(move |inner| {
            let created_mapping =
                Bidirectional::new(&inner.partition, HashCreatedMapping, CreatedHashMapping);
            let identifier_mapping = HashIdentifierMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            if let Err(MappingAlreadyExists) =
                created_mapping.insert_unique(&mut tx, hash.clone(), Created(created_key))?
            {
                tx.rollback();
                return Ok(Err(HashAlreadyExists));
            }

            identifier_mapping.insert(&mut tx, hash, identifier);

            tx.commit()?;

            Ok(Ok(()))
        })
        .await
    }

    async fn update_identifier(&self, hash: Hash, identifier: &Arc<str>) -> Result<(), RepoError> {
        let identifier = Identifier(Arc::clone(identifier));

        self.op(move |inner| {
            let identifier_mapping = HashIdentifierMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            identifier_mapping.insert(&mut tx, hash, identifier);

            tx.commit()?;

            Ok(())
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn identifier(&self, hash: Hash) -> Result<Option<Arc<str>>, RepoError> {
        self.op(move |inner| {
            let identifier_mapping = HashIdentifierMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let opt = identifier_mapping.get(&tx, hash)?;

            Ok(opt.map(|Identifier(id)| id))
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn relate_blurhash(&self, hash: Hash, blurhash: Arc<str>) -> Result<(), RepoError> {
        self.op(move |inner| {
            let blurhash_mapping = HashBlurhashMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            blurhash_mapping.insert(&mut tx, hash, Blurhash(blurhash));

            tx.commit()?;

            Ok(())
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn blurhash(&self, hash: Hash) -> Result<Option<Arc<str>>, RepoError> {
        self.op(move |inner| {
            let blurhash_mapping = HashBlurhashMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let opt = blurhash_mapping.get(&tx, hash)?;

            Ok(opt.map(|Blurhash(bh)| bh))
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn relate_motion_identifier(
        &self,
        hash: Hash,
        identifier: &Arc<str>,
    ) -> Result<(), RepoError> {
        let motion_identifier = MotionIdentifier::new(identifier);

        self.op(move |inner| {
            let motion_identifier_mapping = HashMotionIdentifierMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            motion_identifier_mapping.insert(&mut tx, hash, motion_identifier);

            tx.commit()?;

            Ok(())
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn motion_identifier(&self, hash: Hash) -> Result<Option<Arc<str>>, RepoError> {
        self.op(move |inner| {
            let motion_identifier_mapping = HashMotionIdentifierMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let opt = motion_identifier_mapping.get(&tx, hash)?;

            Ok(opt.map(|MotionIdentifier(Identifier(mi))| mi))
        })
        .await
    }

    #[tracing::instrument(skip(self))]
    async fn cleanup_hash(&self, hash: Hash) -> Result<(), RepoError> {
        self.op(move |inner| {
            let created_mapping =
                Bidirectional::new(&inner.partition, HashCreatedMapping, CreatedHashMapping);
            let identifier_mapping = HashIdentifierMapping.tree(&inner.partition);
            let blurhash_mapping = HashBlurhashMapping.tree(&inner.partition);
            let motion_identifier_mapping = HashMotionIdentifierMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            created_mapping.remove(&mut tx, hash.clone())?;
            identifier_mapping.remove(&mut tx, hash.clone());
            blurhash_mapping.remove(&mut tx, hash.clone());
            motion_identifier_mapping.remove(&mut tx, hash);

            tx.commit()?;

            Ok(())
        })
        .await
    }
}

fn variant_notification_key(hash: &Hash, variant: &str) -> Arc<str> {
    Arc::from(format!("{}{variant}", hash.to_base64()))
}

#[async_trait::async_trait(?Send)]
impl VariantRepo for FjallRepo {
    #[tracing::instrument(level = "trace", skip(self))]
    async fn claim_variant_processing_rights(
        &self,
        hash: Hash,
        variant: String,
    ) -> Result<Result<(), NotificationEntry>, RepoError> {
        let key = (hash.clone(), variant.clone());
        let now = time::OffsetDateTime::now_utc();
        let entry = self
            .inner
            .variant_notifications
            .register_interest(variant_notification_key(&hash, &variant));

        match self.inner.variant_process_map.entry(key.clone()) {
            dashmap::mapref::entry::Entry::Occupied(mut occupied_entry) => {
                if occupied_entry
                    .get()
                    .saturating_add(time::Duration::minutes(2))
                    > now
                {
                    return Ok(Err(entry));
                }

                occupied_entry.insert(now);
            }
            dashmap::mapref::entry::Entry::Vacant(vacant_entry) => {
                vacant_entry.insert(now);
            }
        }

        if self.variant_identifier(hash, variant).await?.is_some() {
            self.inner.variant_process_map.remove(&key);
            return Ok(Err(entry));
        }

        Ok(Ok(()))
    }

    async fn variant_waiter(
        &self,
        hash: Hash,
        variant: String,
    ) -> Result<NotificationEntry, RepoError> {
        let entry = self
            .inner
            .variant_notifications
            .register_interest(variant_notification_key(&hash, &variant));

        Ok(entry)
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn variant_heartbeat(&self, hash: Hash, variant: String) -> Result<(), RepoError> {
        let key = (hash, variant);
        let now = time::OffsetDateTime::now_utc();

        if let dashmap::mapref::entry::Entry::Occupied(mut occupied_entry) =
            self.inner.variant_process_map.entry(key)
        {
            occupied_entry.insert(now);
        }

        Ok(())
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn notify_variant(&self, hash: Hash, variant: String) -> Result<(), RepoError> {
        let key = (hash.clone(), variant.clone());
        self.inner.variant_process_map.remove(&key);

        let key = variant_notification_key(&hash, &variant);
        self.inner.variant_notifications.notify(&key);

        Ok(())
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn relate_variant_identifier(
        &self,
        hash: Hash,
        variant: String,
        identifier: &Arc<str>,
    ) -> Result<Result<(), VariantAlreadyExists>, RepoError> {
        let key = HashVariant(hash, variant);
        let value = Identifier(Arc::clone(identifier));

        self.op(move |inner| {
            let variant_identifier_mapping = HashVariantIdentifierMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            if variant_identifier_mapping.contains_key(&mut tx, key.clone())? {
                tx.rollback();
                return Ok(Err(VariantAlreadyExists));
            }

            variant_identifier_mapping.insert(&mut tx, key, value);

            tx.commit()?;

            Ok(Ok(()))
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn variant_identifier(
        &self,
        hash: Hash,
        variant: String,
    ) -> Result<Option<Arc<str>>, RepoError> {
        self.op(move |inner| {
            let mapping = HashVariantIdentifierMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let opt = mapping.get(&tx, HashVariant(hash, variant))?;

            Ok(opt.map(|Identifier(id)| id))
        })
        .await
    }

    #[tracing::instrument(level = "debug", skip(self))]
    async fn variants(&self, hash: Hash) -> Result<Vec<(String, Arc<str>)>, RepoError> {
        self.op(move |inner| {
            let mapping = HashVariantIdentifierMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            mapping
                .prefix(&tx, Some(HashVariant(hash, String::new())))
                .map(|res| {
                    res.map(|(HashVariant(_, variant), Identifier(identifier))| {
                        (variant, identifier)
                    })
                })
                .collect()
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn remove_variant(&self, hash: Hash, variant: String) -> Result<(), RepoError> {
        self.op(move |inner| {
            let mapping = HashVariantIdentifierMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            mapping.remove(&mut tx, HashVariant(hash, variant));

            tx.commit()?;

            Ok(())
        })
        .await
    }
}

#[async_trait::async_trait(?Send)]
impl AliasRepo for FjallRepo {
    #[tracing::instrument(level = "trace", skip(self))]
    async fn create_alias(
        &self,
        alias: &Alias,
        delete_token: &DeleteToken,
        hash: Hash,
    ) -> Result<Result<(), AliasAlreadyExists>, RepoError> {
        let alias = alias.clone();
        let delete_token = delete_token.clone();

        self.op(move |inner| {
            let delete_token_mapping = AliasDeleteTokenMapping.tree(&inner.partition);
            let alias_hash_mapping = AliasHashMapping.tree(&inner.partition);
            let hash_alias_mapping = HashAliasMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            if alias_hash_mapping.contains_key(&mut tx, alias.clone())? {
                tx.rollback();
                return Ok(Err(AliasAlreadyExists));
            }

            alias_hash_mapping.insert(&mut tx, alias.clone(), hash.clone());
            hash_alias_mapping.insert(&mut tx, HashAlias(hash, alias.clone()), Empty);
            delete_token_mapping.insert(&mut tx, alias, delete_token);

            tx.commit()?;

            Ok(Ok(()))
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn delete_token(&self, alias: &Alias) -> Result<Option<DeleteToken>, RepoError> {
        let alias = alias.clone();

        self.op(move |inner| {
            let delete_token_mapping = AliasDeleteTokenMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let opt = delete_token_mapping.get(&tx, alias)?;

            Ok(opt)
        })
        .await
    }

    #[tracing::instrument(level = "trace", skip(self))]
    async fn hash(&self, alias: &Alias) -> Result<Option<Hash>, RepoError> {
        let alias = alias.clone();

        self.op(move |inner| {
            let mapping = AliasHashMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            let opt = mapping.get(&tx, alias)?;

            Ok(opt)
        })
        .await
    }

    #[tracing::instrument(skip_all)]
    async fn aliases_for_hash(&self, hash: Hash) -> Result<Vec<Alias>, RepoError> {
        self.op(move |inner| {
            let hash_alias_mapping = HashAliasMapping.tree(&inner.partition);

            let tx = inner.db.read_tx();

            hash_alias_mapping
                .prefix(
                    &tx,
                    Some(HashAlias(hash, Alias::generate(".whatevs".into()))),
                )
                .map(|res| res.map(|(HashAlias(_hash, alias), Empty)| alias))
                .collect()
        })
        .await
    }

    #[tracing::instrument(skip(self))]
    async fn cleanup_alias(&self, alias: &Alias) -> Result<(), RepoError> {
        let alias = alias.clone();

        self.op(move |inner| {
            let hash_alias_mapping = HashAliasMapping.tree(&inner.partition);
            let alias_hash_mapping = AliasHashMapping.tree(&inner.partition);
            let delete_token_mapping = AliasDeleteTokenMapping.tree(&inner.partition);

            let mut tx = inner.write_tx();

            if let Some(hash) = alias_hash_mapping.write_get(&mut tx, alias.clone())? {
                alias_hash_mapping.remove(&mut tx, alias.clone());
                hash_alias_mapping.remove(&mut tx, HashAlias(hash, alias.clone()));
            }

            delete_token_mapping.remove(&mut tx, alias);

            tx.commit()?;

            Ok(())
        })
        .await
    }
}

impl std::fmt::Debug for FjallRepo {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("FjallRepo").finish()
    }
}
