use fjall::{ReadTransaction, TransactionalPartitionHandle, WriteTransaction};
use std::ops::Bound;

use super::tree::Tree;

pub(super) struct MappingAlreadyExists;

pub(super) struct Bidirectional<'a, Forward, Inverse> {
    forward: Tree<'a, Forward>,
    inverse: Tree<'a, Inverse>,
}

impl<'a, Forward, Inverse> Bidirectional<'a, Forward, Inverse> {
    pub(super) fn new(
        partition: &'a TransactionalPartitionHandle,
        forward: Forward,
        inverse: Inverse,
    ) -> Self
    where
        Forward: super::tree::Mapping,
        Inverse: super::tree::Mapping<
            Error = Forward::Error,
            Key = Forward::Value,
            Value = Forward::Key,
        >,
    {
        Self {
            forward: forward.tree(partition),
            inverse: inverse.tree(partition),
        }
    }

    pub(super) fn len(
        &self,
        transaction: &ReadTransaction,
        prefix_key: Option<Forward::Key>,
    ) -> fjall::Result<usize>
    where
        Forward: super::tree::Mapping,
        Inverse: super::tree::Mapping<
            Error = Forward::Error,
            Key = Forward::Value,
            Value = Forward::Key,
        >,
    {
        self.forward.len(transaction, prefix_key)
    }

    pub(super) fn insert_unique(
        &self,
        transaction: &mut WriteTransaction,
        key: Forward::Key,
        value: Inverse::Key,
    ) -> fjall::Result<Result<(), MappingAlreadyExists>>
    where
        Forward: super::tree::Mapping,
        Inverse: super::tree::Mapping<
            Error = Forward::Error,
            Key = Forward::Value,
            Value = Forward::Key,
        >,
        Forward::Key: Clone,
        Inverse::Key: Clone,
    {
        if self.forward.contains_key(transaction, key.clone())? {
            return Ok(Err(MappingAlreadyExists));
        }

        self.forward.insert(transaction, key.clone(), value.clone());
        self.inverse.insert(transaction, value, key);

        Ok(Ok(()))
    }

    pub(super) fn overwrite(
        &self,
        transaction: &mut WriteTransaction,
        key: Forward::Key,
        value: Inverse::Key,
    ) -> Result<(), Forward::Error>
    where
        Forward: super::tree::Mapping,
        Inverse: super::tree::Mapping<
            Error = Forward::Error,
            Key = Forward::Value,
            Value = Forward::Key,
        >,
        Forward::Key: Clone,
        Inverse::Key: Clone,
    {
        if let Some(previous_value) = self.forward.write_get(transaction, key.clone())? {
            self.inverse.remove(transaction, previous_value);
        }

        self.forward.insert(transaction, key.clone(), value.clone());
        self.inverse.insert(transaction, value, key);

        Ok(())
    }

    pub(super) fn get(
        &self,
        transaction: &ReadTransaction,
        key: Forward::Key,
    ) -> Result<Option<Forward::Value>, Forward::Error>
    where
        Forward: super::tree::Mapping,
        Inverse: super::tree::Mapping<
            Error = Forward::Error,
            Key = Forward::Value,
            Value = Forward::Key,
        >,
    {
        self.forward.get(transaction, key)
    }

    pub(super) fn range_by_value<'b>(
        &'b self,
        transaction: &'b ReadTransaction,
        range: (Bound<Inverse::Key>, Bound<Inverse::Key>),
    ) -> impl DoubleEndedIterator<Item = Result<(Forward::Key, Inverse::Key), Forward::Error>> + 'static
    where
        Forward: super::tree::Mapping,
        Inverse: super::tree::Mapping<
            Error = Forward::Error,
            Key = Forward::Value,
            Value = Forward::Key,
        >,
        Forward::Key: 'static,
        Inverse::Key: 'static,
    {
        self.inverse
            .range(transaction, range)
            .map(|res| res.map(|(v, k)| (k, v)))
    }

    pub(super) fn remove_by_value(
        &self,
        transaction: &mut WriteTransaction,
        value: Inverse::Key,
    ) -> Result<(), Forward::Error>
    where
        Forward: super::tree::Mapping,
        Inverse: super::tree::Mapping<
            Error = Forward::Error,
            Key = Forward::Value,
            Value = Forward::Key,
        >,
        Inverse::Key: Clone,
    {
        if let Some(key) = self.inverse.write_get(transaction, value.clone())? {
            self.forward.remove(transaction, key);
            self.inverse.remove(transaction, value);
        }

        Ok(())
    }

    pub(super) fn remove(
        &self,
        transaction: &mut WriteTransaction,
        key: Forward::Key,
    ) -> Result<(), Forward::Error>
    where
        Forward: super::tree::Mapping,
        Inverse: super::tree::Mapping<
            Error = Forward::Error,
            Key = Forward::Value,
            Value = Forward::Key,
        >,
        Forward::Key: Clone,
    {
        if let Some(value) = self.forward.write_get(transaction, key.clone())? {
            self.inverse.remove(transaction, value);
            self.forward.remove(transaction, key);
        }

        Ok(())
    }
}
