use std::{array::TryFromSliceError, borrow::Cow};

use crate::{
    path_gen::{PathItem, PathNode},
    repo::{Alias, Hash},
};

pub(super) struct HashAlias(pub Hash, pub Alias);

#[derive(Debug, thiserror::Error)]
pub(crate) enum HashAliasError {
    #[error("Error parsing Hash length")]
    Slice(#[from] TryFromSliceError),

    #[error("Error parsing hash")]
    Hash,

    #[error("Error parsing alias")]
    Alias,
}

impl<'a> PathItem<'a> for HashAlias {
    type Error = HashAliasError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        let hash_bytes = self.0.to_bytes();
        let alias_bytes = self.1.to_bytes();

        let hash_len = hash_bytes.len() as u64;

        let mut bytes = hash_len.to_be_bytes().to_vec();
        bytes.extend(hash_bytes);
        bytes.extend(alias_bytes);

        Cow::Owned(bytes)
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        let len_bytes = bytes[0..8].try_into()?;
        let len = u64::from_be_bytes(len_bytes) as usize;

        let bytes = &bytes[8..];

        let hash = Hash::from_bytes(&bytes[..len]).ok_or(HashAliasError::Hash)?;

        let alias = Alias::from_slice(&bytes[len..]).ok_or(HashAliasError::Alias)?;

        Ok(Self(hash, alias))
    }
}

impl<'a> PathNode<'a> for HashAlias {
    const NAME: &'static [u8] = b"hash-alias";
}

impl std::fmt::Debug for HashAlias {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.to_bytes().len().fmt(f)?;
        self.0.fmt(f)?;
        self.1.fmt(f)
    }
}
