use std::{array::TryFromSliceError, borrow::Cow, num::TryFromIntError, str::Utf8Error};

use crate::{
    path_gen::{PathItem, PathNode},
    repo::Hash,
};

#[derive(Clone)]
pub(super) struct HashVariant(pub Hash, pub String);

#[derive(Debug, thiserror::Error)]
pub(crate) enum HashVariantError {
    #[error("Invalid hash length")]
    HashLength(#[from] TryFromSliceError),

    #[error("Hash length too large")]
    HashSize(#[from] TryFromIntError),

    #[error("Failed to parse Hash")]
    Parse,

    #[error("Failed to parse variant")]
    Utf8(#[from] Utf8Error),
}

impl<'a> PathItem<'a> for HashVariant {
    type Error = HashVariantError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        let hash = self.0.to_bytes();
        let hash_len = u64::try_from(hash.len()).expect("hash is reasonable length");
        let variant = self.1.as_bytes();

        let mut out = Vec::with_capacity(8 + hash.len() + variant.len());

        out.extend(hash_len.to_be_bytes());
        out.extend(hash);
        out.extend(variant);

        Cow::Owned(out)
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        let hash_len_bytes = bytes[0..8].try_into()?;
        let hash_len = u64::from_be_bytes(hash_len_bytes);

        let hash_len: usize = hash_len.try_into()?;

        let bytes = &bytes[8..];

        let hash = Hash::from_bytes(&bytes[..hash_len]).ok_or(HashVariantError::Parse)?;

        let bytes = &bytes[hash_len..];

        let variant = std::str::from_utf8(bytes)?;

        Ok(Self(hash, String::from(variant)))
    }
}

impl<'a> PathNode<'a> for HashVariant {
    const NAME: &'static [u8] = b"hash-variant";
}

impl std::fmt::Debug for HashVariant {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.to_bytes().len().fmt(f)?;
        self.0.fmt(f)?;
        self.1.fmt(f)
    }
}
