use crate::path_gen::{PathItem, PathNode};
use std::borrow::Cow;

use super::FjallError;

pub(super) struct HealthzKey;
pub(super) struct Healthz(pub u64);

impl<'a> PathItem<'a> for HealthzKey {
    type Error = FjallError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        Cow::Borrowed(b"healthz")
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        if &bytes[..] != b"healthz" {
            Err(FjallError::HealthzKey)
        } else {
            Ok(HealthzKey)
        }
    }
}

impl<'a> PathItem<'a> for Healthz {
    type Error = FjallError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        Cow::Owned(self.0.to_be_bytes().to_vec())
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        bytes[0..8]
            .try_into()
            .map_err(FjallError::Healthz)
            .map(u64::from_be_bytes)
            .map(Self)
    }
}

impl<'a> PathNode<'a> for HealthzKey {
    const NAME: &'static [u8] = b"healthz-key";
}

impl<'a> PathNode<'a> for Healthz {
    const NAME: &'static [u8] = b"healthz";
}

impl std::fmt::Debug for HealthzKey {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("healthz-key")
    }
}

impl std::fmt::Debug for Healthz {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}
