use crate::path_gen::{PathItem, PathNode};
use std::{borrow::Cow, sync::Arc};

use super::FjallError;

pub(super) struct Identifier(pub Arc<str>);
pub(super) struct MigratedIdentifier(pub Identifier);
pub(super) struct MotionIdentifier(pub Identifier);

impl Identifier {
    pub(super) fn new(identifier: &Arc<str>) -> Self {
        Self(Arc::clone(identifier))
    }
}

impl MigratedIdentifier {
    pub(super) fn new(identifier: &Arc<str>) -> Self {
        Self(Identifier::new(identifier))
    }
}

impl MotionIdentifier {
    pub(super) fn new(identifier: &Arc<str>) -> Self {
        Self(Identifier::new(identifier))
    }
}

impl<'a> PathItem<'a> for Identifier {
    type Error = FjallError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        Cow::Borrowed(self.0.as_bytes())
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        match bytes {
            Cow::Borrowed(s) => std::str::from_utf8(s)
                .map_err(FjallError::Identifier)
                .map(String::from)
                .map(Arc::from)
                .map(Identifier),
            Cow::Owned(s) => String::from_utf8(s)
                .map_err(|e| e.utf8_error())
                .map_err(FjallError::Identifier)
                .map(Arc::from)
                .map(Identifier),
        }
    }
}

impl<'a> PathItem<'a> for MigratedIdentifier {
    type Error = FjallError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        self.0.to_bytes()
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        Identifier::parse(bytes).map(Self)
    }
}

impl<'a> PathItem<'a> for MotionIdentifier {
    type Error = FjallError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        self.0.to_bytes()
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        Identifier::parse(bytes).map(Self)
    }
}

impl<'a> PathNode<'a> for Identifier {
    const NAME: &'static [u8] = b"identifier";
}

impl<'a> PathNode<'a> for MigratedIdentifier {
    const NAME: &'static [u8] = b"migrated-identifier";
}

impl<'a> PathNode<'a> for MotionIdentifier {
    const NAME: &'static [u8] = b"motion-identifier";
}

impl std::fmt::Debug for Identifier {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

impl std::fmt::Debug for MigratedIdentifier {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

impl std::fmt::Debug for MotionIdentifier {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}
