use crate::path_gen::{PathItem, PathNode};
use std::borrow::Cow;
use uuid::Uuid;

use super::FjallError;

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
pub(super) enum JobState {
    Pending,
    Running([u8; 24]),
}

impl JobState {
    pub(super) fn running(worker_id: Uuid) -> Self {
        let first_eight = time::OffsetDateTime::now_utc()
            .unix_timestamp()
            .to_be_bytes();

        let next_sixteen = worker_id.into_bytes();

        let mut bytes = [0u8; 24];

        bytes[0..8]
            .iter_mut()
            .zip(&first_eight)
            .for_each(|(dest, src)| *dest = *src);

        bytes[8..24]
            .iter_mut()
            .zip(&next_sixteen)
            .for_each(|(dest, src)| *dest = *src);

        Self::Running(bytes)
    }

    pub(super) const fn is_pending(&self) -> bool {
        matches!(self, Self::Pending)
    }

    pub(super) fn timestamp(&self) -> Option<time::OffsetDateTime> {
        match self {
            Self::Pending => None,
            Self::Running(ref bytes) => {
                let unix_timestamp =
                    i64::from_be_bytes(bytes[0..8].try_into().expect("Verified length"));

                let timestamp = time::OffsetDateTime::from_unix_timestamp(unix_timestamp)
                    .expect("Valid Timestamp");

                Some(timestamp)
            }
        }
    }

    pub(super) fn is_older_than(
        &self,
        duration: time::Duration,
        now: time::OffsetDateTime,
    ) -> bool {
        matches!(self.timestamp(), Some(timestamp) if timestamp.saturating_add(duration) < now)
    }
}

impl<'a> PathItem<'a> for JobState {
    type Error = FjallError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        match self {
            Self::Pending => Cow::Borrowed(b"pend"),
            Self::Running(bytes) => Cow::Borrowed(bytes),
        }
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        if &bytes[..] == b"pend" {
            Ok(Self::Pending)
        } else {
            bytes[..]
                .try_into()
                .map(Self::Running)
                .map_err(FjallError::JobState)
        }
    }
}
impl<'a> PathNode<'a> for JobState {
    const NAME: &'static [u8] = b"state";
}
