use super::{
    tree::Mapping as Map, Accessed, Alias, Blurhash, Created, DeleteToken, DetailsInner, Empty,
    FjallError, Hash, HashAlias, HashVariant, Healthz, HealthzKey, Identifier, Job, JobId,
    JobState, MigratedIdentifier, MotionIdentifier, Proxy, Queue, QueueJobKey, RetryCount, Setting,
    SettingsValue, UniqueJob, UploadId, UploadValue, Variant,
};
use crate::path_gen::{PathGen, PathItem};

#[derive(Debug, Clone, Copy)]
pub(super) struct HashCreatedMapping;
pub(super) struct CreatedHashMapping;
pub(super) struct HealthzMapping;
pub(super) struct ProxyAliasMapping;
pub(super) struct AliasProxyMapping;
pub(super) struct AliasAccessMapping;
pub(super) struct AccessAliasMapping;
pub(super) struct VariantAccessMapping;
pub(super) struct AccessVariantMapping;
pub(super) struct UploadMapping;
pub(super) struct UniqueJobQueueMapping;
pub(super) struct QueueUniqueJobMapping;
pub(super) struct JobMapping;
pub(super) struct JobStateMapping;
pub(super) struct JobRetryMapping;
pub(super) struct SettingsMapping;
pub(super) struct IdentifierDetailsMapping;
pub(super) struct MigratedIdentifierMapping;
pub(super) struct HashIdentifierMapping;
pub(super) struct HashMotionIdentifierMapping;
pub(super) struct HashBlurhashMapping;
pub(super) struct HashVariantIdentifierMapping;
pub(super) struct AliasHashMapping;
pub(super) struct HashAliasMapping;
pub(super) struct AliasDeleteTokenMapping;

enum FjallDebug {
    HashCreated(HashCreatedMapping),
    CreatedHash(CreatedHashMapping),
    Healthz(HealthzMapping),
    ProxyAlias(ProxyAliasMapping),
    AliasProxy(AliasProxyMapping),
    AliasAccess(AliasAccessMapping),
    AccessAlias(AccessAliasMapping),
    VariantAccess(VariantAccessMapping),
    AccessVariant(AccessVariantMapping),
    Upload(UploadMapping),
    UniqueJobQueue(UniqueJobQueueMapping),
    QueueUniqueJob(QueueUniqueJobMapping),
    Job(JobMapping),
    JobState(JobStateMapping),
    JobRetry(JobRetryMapping),
    Settings(SettingsMapping),
    IdentifierDetails(IdentifierDetailsMapping),
    MigratedIdentifier(MigratedIdentifierMapping),
    HashIdentifier(HashIdentifierMapping),
    HashMotionIdentifier(HashMotionIdentifierMapping),
    HashBlurhash(HashBlurhashMapping),
    HashVariantIdentifier(HashVariantIdentifierMapping),
    AliasHash(AliasHashMapping),
    HashAlias(HashAliasMapping),
    AliasDeleteToken(AliasDeleteTokenMapping),
}

const MAPPINGS: &[FjallDebug] = &[
    FjallDebug::HashCreated(HashCreatedMapping),
    FjallDebug::CreatedHash(CreatedHashMapping),
    FjallDebug::Healthz(HealthzMapping),
    FjallDebug::ProxyAlias(ProxyAliasMapping),
    FjallDebug::AliasProxy(AliasProxyMapping),
    FjallDebug::AliasAccess(AliasAccessMapping),
    FjallDebug::AccessAlias(AccessAliasMapping),
    FjallDebug::VariantAccess(VariantAccessMapping),
    FjallDebug::AccessVariant(AccessVariantMapping),
    FjallDebug::Upload(UploadMapping),
    FjallDebug::UniqueJobQueue(UniqueJobQueueMapping),
    FjallDebug::QueueUniqueJob(QueueUniqueJobMapping),
    FjallDebug::Job(JobMapping),
    FjallDebug::JobState(JobStateMapping),
    FjallDebug::JobRetry(JobRetryMapping),
    FjallDebug::Settings(SettingsMapping),
    FjallDebug::IdentifierDetails(IdentifierDetailsMapping),
    FjallDebug::MigratedIdentifier(MigratedIdentifierMapping),
    FjallDebug::HashIdentifier(HashIdentifierMapping),
    FjallDebug::HashMotionIdentifier(HashMotionIdentifierMapping),
    FjallDebug::HashBlurhash(HashBlurhashMapping),
    FjallDebug::HashVariantIdentifier(HashVariantIdentifierMapping),
    FjallDebug::AliasHash(AliasHashMapping),
    FjallDebug::HashAlias(HashAliasMapping),
    FjallDebug::AliasDeleteToken(AliasDeleteTokenMapping),
];

pub(super) struct RowDebugger<'a>(pub &'a [u8], pub &'a [u8]);

impl<'a> std::fmt::Debug for RowDebugger<'a> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        debug_key_value(self.0, self.1, f)
    }
}

fn debug_key_value(key: &[u8], value: &[u8], f: &mut std::fmt::Formatter) -> std::fmt::Result {
    use FjallDebug::*;

    write!(f, "(")?;

    for mapping in MAPPINGS {
        match mapping {
            HashCreated(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            CreatedHash(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            Healthz(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            ProxyAlias(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            AliasProxy(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            AliasAccess(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            AccessAlias(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            VariantAccess(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            AccessVariant(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            Upload(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            UniqueJobQueue(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            QueueUniqueJob(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            Job(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            JobState(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            JobRetry(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            Settings(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            IdentifierDetails(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            MigratedIdentifier(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            HashIdentifier(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            HashMotionIdentifier(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            HashBlurhash(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            HashVariantIdentifier(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            AliasHash(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            HashAlias(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
            AliasDeleteToken(m) => {
                if let Some(res) = try_debug_key_value(m, key, value, f) {
                    return res;
                }
            }
        }
    }

    write!(
        f,
        "Unknown Key {}): <{}>",
        String::from_utf8_lossy(key),
        String::from_utf8_lossy(value)
    )
}

fn try_debug_key_value<M>(
    _mapping: &M,
    key: &[u8],
    value: &[u8],
    f: &mut std::fmt::Formatter,
) -> Option<std::fmt::Result>
where
    M: Map,
{
    let key = M::debug_key_path(key, f)?;

    if key.is_err() {
        return Some(key);
    }

    match M::Value::parse(std::borrow::Cow::Owned(value.to_vec())) {
        Ok(value) => Some(write!(f, "): <{value:?}>")),
        Err(e) => Some(write!(f, "): <{e}>")),
    }
}

impl Map for HashCreatedMapping {
    type Error = FjallError;
    type Key = Hash;
    type Value = Created;
}

impl Map for CreatedHashMapping {
    type Error = FjallError;
    type Key = Created;
    type Value = Hash;
}

impl Map for HealthzMapping {
    type Error = FjallError;
    type Key = HealthzKey;
    type Value = Healthz;

    fn key_path(&self, _key: Self::Key) -> Vec<u8> {
        PathGen.prefix::<Healthz>().to_bytes()
    }

    fn key_prefix(&self, _key: Option<Self::Key>) -> Vec<u8> {
        PathGen.prefix::<Healthz>().to_bytes()
    }

    fn parse_key_path(key_path_bytes: &[u8]) -> Result<Self::Key, Self::Error> {
        PathGen::parser()
            .prefix::<Self::Value>()
            .parse::<Self::Error>(key_path_bytes)?;

        Ok(HealthzKey)
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser().prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }
}

impl Map for ProxyAliasMapping {
    type Error = FjallError;
    type Key = Proxy;
    type Value = Alias;
}

impl Map for AliasProxyMapping {
    type Error = FjallError;
    type Key = Alias;
    type Value = Proxy;
}

impl Map for AliasAccessMapping {
    type Error = FjallError;
    type Key = Alias;
    type Value = Accessed;
}

impl Map for AccessAliasMapping {
    type Error = FjallError;
    type Key = Accessed;
    type Value = Alias;
}

impl Map for VariantAccessMapping {
    type Error = FjallError;
    type Key = HashVariant;
    type Value = Accessed;

    fn key_path(&self, HashVariant(hash, variant): Self::Key) -> Vec<u8> {
        PathGen
            .push(hash)
            .push(Variant(variant))
            .prefix::<Self::Value>()
            .to_bytes()
    }

    fn key_prefix(&self, key: Option<Self::Key>) -> Vec<u8> {
        match key {
            // variants for hash
            Some(HashVariant(hash, _variatn)) => PathGen.push(hash).prefix::<Variant>().to_bytes(),
            // All variants
            None => PathGen.prefix::<Hash>().to_bytes(),
        }
    }

    fn parse_key_path(key_path_bytes: &[u8]) -> Result<Self::Key, Self::Error> {
        let parser = PathGen::parser()
            .push::<Hash>()
            .push::<Variant>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes)?;

        let (root, Variant(variant)) = key_path.pop().split();
        let (_root, hash) = root.split();

        Ok(HashVariant(hash, variant))
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser()
            .push::<Hash>()
            .push::<Variant>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }
}

impl Map for AccessVariantMapping {
    type Error = FjallError;
    type Key = Accessed;
    type Value = HashVariant;
}

impl Map for UploadMapping {
    type Error = FjallError;
    type Key = UploadId;
    type Value = UploadValue;
}

impl Map for UniqueJobQueueMapping {
    type Error = FjallError;
    type Key = UniqueJob;
    type Value = QueueJobKey;
}

impl Map for QueueUniqueJobMapping {
    type Error = FjallError;
    type Key = QueueJobKey;
    type Value = UniqueJob;

    fn key_path(&self, QueueJobKey(queue_name, job_id): Self::Key) -> Vec<u8> {
        PathGen
            .push(Queue(queue_name))
            .push(job_id)
            .prefix::<Self::Value>()
            .to_bytes()
    }

    fn key_prefix(&self, key: Option<Self::Key>) -> Vec<u8> {
        match key {
            // Jobs in provided queue
            Some(QueueJobKey(queue_name, _job_id)) => {
                PathGen.push(Queue(queue_name)).prefix::<JobId>().to_bytes()
            }

            // All jobs
            None => PathGen.prefix::<Queue<'static>>().to_bytes(),
        }
    }

    fn parse_key_path(key_path_bytes: &[u8]) -> Result<Self::Key, Self::Error> {
        let parser = PathGen::parser()
            .push::<Queue>()
            .push::<JobId>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes)?;

        let (root, job_id) = key_path.pop().split();
        let (_root, Queue(queue_name)) = root.split();

        Ok(QueueJobKey::owned(queue_name.into_owned(), job_id))
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser()
            .push::<Queue>()
            .push::<JobId>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }
}

impl Map for JobMapping {
    type Error = FjallError;
    type Key = QueueJobKey;
    type Value = Job;

    fn key_path(&self, QueueJobKey(queue_name, job_id): Self::Key) -> Vec<u8> {
        PathGen
            .push(Queue(queue_name))
            .push(job_id)
            .prefix::<Self::Value>()
            .to_bytes()
    }

    fn key_prefix(&self, key: Option<Self::Key>) -> Vec<u8> {
        match key {
            // Jobs in specific queue
            Some(QueueJobKey(queue_name, _job_id)) => {
                PathGen.push(Queue(queue_name)).prefix::<JobId>().to_bytes()
            }

            // All jobs
            None => PathGen.prefix::<Queue<'static>>().to_bytes(),
        }
    }

    fn parse_key_path(value_path: &[u8]) -> Result<Self::Key, Self::Error> {
        let parser = PathGen::parser()
            .push::<Queue>()
            .push::<JobId>()
            .prefix::<Self::Value>();

        let path = parser.parse::<Self::Error>(value_path)?;

        let (root, job_id) = path.pop().split();
        let (_root, Queue(queue_name)) = root.split();

        Ok(QueueJobKey::owned(queue_name.into_owned(), job_id))
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser()
            .push::<Queue>()
            .push::<JobId>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }
}

impl Map for JobStateMapping {
    type Error = FjallError;
    type Key = QueueJobKey;
    type Value = JobState;

    fn key_path(&self, QueueJobKey(queue_name, job_id): Self::Key) -> Vec<u8> {
        PathGen
            .push(Queue(queue_name))
            .push(job_id)
            .prefix::<Self::Value>()
            .to_bytes()
    }

    fn key_prefix(&self, key: Option<Self::Key>) -> Vec<u8> {
        match key {
            // Jobs in given queue
            Some(QueueJobKey(queue_name, _job_id)) => {
                PathGen.push(Queue(queue_name)).prefix::<JobId>().to_bytes()
            }

            // All jobs
            None => PathGen.prefix::<Queue<'static>>().to_bytes(),
        }
    }

    fn parse_key_path(value_path: &[u8]) -> Result<Self::Key, Self::Error> {
        let parser = PathGen::parser()
            .push::<Queue>()
            .push::<JobId>()
            .prefix::<Self::Value>();

        let path = parser.parse::<Self::Error>(value_path)?;

        let (root, job_id) = path.pop().split();
        let (_root, Queue(queue_name)) = root.split();

        Ok(QueueJobKey::owned(queue_name.into_owned(), job_id))
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser()
            .push::<Queue>()
            .push::<JobId>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }
}

impl Map for JobRetryMapping {
    type Error = FjallError;
    type Key = QueueJobKey;
    type Value = RetryCount;

    fn key_path(&self, QueueJobKey(queue_name, job_id): Self::Key) -> Vec<u8> {
        PathGen
            .push(Queue(queue_name))
            .push(job_id)
            .prefix::<Self::Value>()
            .to_bytes()
    }

    fn key_prefix(&self, key: Option<Self::Key>) -> Vec<u8> {
        match key {
            // Jobs in given queue
            Some(QueueJobKey(queue_name, _job_id)) => {
                PathGen.push(Queue(queue_name)).prefix::<JobId>().to_bytes()
            }

            // All jobs
            None => PathGen.prefix::<Queue<'static>>().to_bytes(),
        }
    }

    fn parse_key_path(value_path: &[u8]) -> Result<Self::Key, Self::Error> {
        let parser = PathGen::parser()
            .push::<Queue>()
            .push::<JobId>()
            .prefix::<Self::Value>();

        let path = parser.parse::<Self::Error>(value_path)?;

        let (root, job_id) = path.pop().split();
        let (_root, Queue(queue_name)) = root.split();

        Ok(QueueJobKey::owned(queue_name.into_owned(), job_id))
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser()
            .push::<Queue>()
            .push::<JobId>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }
}

impl Map for SettingsMapping {
    type Error = FjallError;
    type Key = Setting;
    type Value = SettingsValue;
}

impl Map for IdentifierDetailsMapping {
    type Error = FjallError;
    type Key = Identifier;
    type Value = DetailsInner;
}

impl Map for MigratedIdentifierMapping {
    type Error = FjallError;
    type Key = MigratedIdentifier;
    type Value = Identifier;
}

impl Map for HashIdentifierMapping {
    type Error = FjallError;
    type Key = Hash;
    type Value = Identifier;
}

impl Map for HashMotionIdentifierMapping {
    type Error = FjallError;
    type Key = Hash;
    type Value = MotionIdentifier;
}

impl Map for HashBlurhashMapping {
    type Error = FjallError;
    type Key = Hash;
    type Value = Blurhash;
}

impl Map for HashVariantIdentifierMapping {
    type Error = FjallError;
    type Key = HashVariant;
    type Value = Identifier;

    fn key_path(&self, HashVariant(hash, variant): Self::Key) -> Vec<u8> {
        PathGen
            .push(hash)
            .push(Variant(variant))
            .prefix::<Identifier>()
            .to_bytes()
    }

    fn key_prefix(&self, key: Option<Self::Key>) -> Vec<u8> {
        match key {
            // Variant Identifiers for given hash
            Some(HashVariant(hash, _variant)) => PathGen.push(hash).prefix::<Variant>().to_bytes(),

            // All Variant identifiers
            None => PathGen.prefix::<Hash>().to_bytes(),
        }
    }

    fn parse_key_path(key_path_bytes: &[u8]) -> Result<Self::Key, Self::Error> {
        let parser = PathGen::parser()
            .push::<Hash>()
            .push::<Variant>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes)?;

        let (root, Variant(variant)) = key_path.pop().split();
        let (_root, hash) = root.split();

        Ok(HashVariant(hash, variant))
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser()
            .push::<Hash>()
            .push::<Variant>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }
}

impl Map for AliasHashMapping {
    type Error = FjallError;
    type Key = Alias;
    type Value = Hash;
}

impl Map for HashAliasMapping {
    type Error = FjallError;
    type Key = HashAlias;
    type Value = Empty;

    fn key_path(&self, HashAlias(hash, alias): Self::Key) -> Vec<u8> {
        PathGen.push(hash).push(alias).prefix::<Empty>().to_bytes()
    }

    fn key_prefix(&self, key: Option<Self::Key>) -> Vec<u8> {
        match key {
            // Aliases for given hash
            Some(HashAlias(hash, _alias)) => PathGen.push(hash).prefix::<Alias>().to_bytes(),

            // All aliases
            None => PathGen.prefix::<Hash>().to_bytes(),
        }
    }

    fn parse_key_path(key_path_bytes: &[u8]) -> Result<Self::Key, Self::Error> {
        let parser = PathGen::parser()
            .push::<Hash>()
            .push::<Alias>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes)?;

        let (root, alias) = key_path.pop().split();

        let (_root, hash) = root.split();

        Ok(HashAlias(hash, alias))
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser()
            .push::<Hash>()
            .push::<Alias>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }
}

impl Map for AliasDeleteTokenMapping {
    type Error = FjallError;
    type Key = Alias;
    type Value = DeleteToken;
}
