use std::{array::TryFromSliceError, borrow::Cow, num::TryFromIntError, str::Utf8Error};

use crate::{
    path_gen::{PathItem, PathNode},
    repo::JobId,
};

#[derive(Clone, PartialEq, Eq, PartialOrd, Ord)]
pub(super) struct QueueJobKey(pub Cow<'static, str>, pub JobId);

impl QueueJobKey {
    pub(super) const fn new(queue_name: &'static str, job_id: JobId) -> Self {
        Self(Cow::Borrowed(queue_name), job_id)
    }

    pub(super) const fn owned(queue: String, job_id: JobId) -> Self {
        Self(Cow::Owned(queue), job_id)
    }

    pub(super) fn prefix(queue_name: &'static str) -> Self {
        Self(Cow::Borrowed(queue_name), JobId::gen())
    }
}

#[derive(Debug, thiserror::Error)]
pub(crate) enum QueueJobKeyError {
    #[error("Invalid queue name length")]
    QueueLength(#[from] TryFromSliceError),

    #[error("Queue name too long")]
    QueueSize(#[from] TryFromIntError),

    #[error("Invalid queue string")]
    Utf8(#[from] Utf8Error),
}

impl<'a> PathItem<'a> for QueueJobKey {
    type Error = QueueJobKeyError;

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        let len = bytes[0..8].try_into()?;
        let len = u64::from_be_bytes(len);
        let len = len as usize;
        let bytes = &bytes[8..];

        let queue_name = std::str::from_utf8(&bytes[..len])?;

        let job_id = bytes[len..].try_into()?;
        let job_id = JobId::from_bytes(job_id);

        Ok(QueueJobKey(Cow::Owned(String::from(queue_name)), job_id))
    }

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        let key_bytes = self.0.as_bytes();
        let mut key = (key_bytes.len() as u64).to_be_bytes().to_vec();
        key.extend(key_bytes);
        key.extend(self.1.as_bytes());

        Cow::Owned(key)
    }
}

impl<'a> PathNode<'a> for QueueJobKey {
    const NAME: &'static [u8] = b"queue-job";
}

impl std::fmt::Debug for QueueJobKey {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.as_bytes().len().fmt(f)?;
        self.0.fmt(f)?;
        self.1.fmt(f)
    }
}
