use crate::path_gen::{PathItem, PathNode};
use std::{borrow::Cow, sync::Arc};

use super::FjallError;

#[derive(Clone)]
pub(super) enum Setting {
    New(&'static str),
    Existing(String),
}

pub(super) struct SettingsValue(pub Arc<[u8]>);

impl<'a> PathItem<'a> for Setting {
    type Error = FjallError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        match self {
            Self::New(queue_name) => Cow::Borrowed(queue_name.as_bytes()),
            Self::Existing(existing) => Cow::Borrowed(existing.as_bytes()),
        }
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        match bytes {
            Cow::Borrowed(s) => std::str::from_utf8(s)
                .map_err(FjallError::Setting)
                .map(String::from)
                .map(Self::Existing),
            Cow::Owned(s) => String::from_utf8(s)
                .map_err(|e| e.utf8_error())
                .map_err(FjallError::Setting)
                .map(Self::Existing),
        }
    }
}

impl<'a> PathItem<'a> for SettingsValue {
    type Error = std::convert::Infallible;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        Cow::Borrowed(&self.0)
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        Ok(Self(Arc::from(bytes.into_owned())))
    }
}

impl<'a> PathNode<'a> for Setting {
    const NAME: &'static [u8] = b"setting";
}

impl<'a> PathNode<'a> for SettingsValue {
    const NAME: &'static [u8] = b"settings-value";
}

impl std::fmt::Debug for Setting {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::New(s) => f.write_str(s),
            Self::Existing(s) => f.write_str(s),
        }
    }
}

impl std::fmt::Debug for SettingsValue {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        hex::encode(&self.0[..]).fmt(f)
    }
}
