use crate::path_gen::{PathItem, PathNode};
use std::{array::TryFromSliceError, borrow::Cow};
use time::{error::ComponentRange, OffsetDateTime};

#[derive(Clone)]
pub(super) struct Timestamp(pub OffsetDateTime, pub Vec<u8>);

#[derive(Clone)]
pub(super) struct Created(pub Timestamp);

#[derive(Clone)]
pub(super) struct Accessed(pub Timestamp);

#[derive(Debug, thiserror::Error)]
pub(crate) enum TimeError {
    #[error("Invalid timestamp length")]
    Bytes(#[from] TryFromSliceError),

    #[error("Invalid timestamp")]
    Time(#[from] ComponentRange),
}

impl<'a> PathItem<'a> for Timestamp {
    type Error = TimeError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        let mut bytes = self.0.unix_timestamp_nanos().to_be_bytes().to_vec();
        bytes.extend_from_slice(&self.1);

        Cow::Owned(bytes)
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        let timestamp_bytes = bytes[0..16].try_into()?;

        let timestamp =
            OffsetDateTime::from_unix_timestamp_nanos(i128::from_be_bytes(timestamp_bytes))?;

        Ok(Timestamp(timestamp, bytes[16..].into()))
    }
}

impl<'a> PathItem<'a> for Created {
    type Error = TimeError;

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        Timestamp::parse(bytes).map(Created)
    }

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        self.0.to_bytes()
    }
}

impl<'a> PathItem<'a> for Accessed {
    type Error = TimeError;

    fn to_bytes(&self) -> Cow<'_, [u8]> {
        self.0.to_bytes()
    }

    fn parse(bytes: Cow<'a, [u8]>) -> Result<Self, Self::Error>
    where
        Self: Sized + 'a,
    {
        Timestamp::parse(bytes).map(Self)
    }
}

impl<'a> PathNode<'a> for Created {
    const NAME: &'static [u8] = b"created";
}

impl<'a> PathNode<'a> for Accessed {
    const NAME: &'static [u8] = b"accessed";
}

impl std::fmt::Debug for Timestamp {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)?;
        hex::encode(&self.1).fmt(f)
    }
}

impl std::fmt::Debug for Created {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

impl std::fmt::Debug for Accessed {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}
