use crate::path_gen::{DeconstructError, PathGen, PathItem, PathNode};
use fjall::{ReadTransaction, TransactionalPartitionHandle, WriteTransaction};
use std::{borrow::Cow, ops::Bound};

pub(super) struct Tree<'a, M> {
    partition: &'a TransactionalPartitionHandle,
    mapping: M,
}

pub(super) trait Mapping {
    type Error: From<DeconstructError>
        + for<'k> From<<Self::Key as PathItem<'k>>::Error>
        + for<'v> From<<Self::Value as PathItem<'v>>::Error>
        + From<fjall::Error>;

    type Key: for<'a> PathNode<'a> + std::fmt::Debug;
    type Value: for<'a> PathNode<'a> + std::fmt::Debug;

    fn key_path(&self, key: Self::Key) -> Vec<u8> {
        PathGen.push(key).prefix::<Self::Value>().to_bytes()
    }

    fn key_prefix(&self, _key: Option<Self::Key>) -> Vec<u8> {
        PathGen.prefix::<Self::Key>().to_bytes()
    }

    fn parse_key_path(key_path_bytes: &[u8]) -> Result<Self::Key, Self::Error> {
        let parser = PathGen::parser()
            .push::<Self::Key>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes)?;

        let (_root, key) = key_path.pop().split();

        Ok(key)
    }

    fn debug_key_path(
        key_path_bytes: &[u8],
        f: &mut std::fmt::Formatter,
    ) -> Option<std::fmt::Result> {
        let parser = PathGen::parser()
            .push::<Self::Key>()
            .prefix::<Self::Value>();

        let key_path = parser.parse::<Self::Error>(key_path_bytes).ok()?;

        Some(write!(f, "{key_path:?}"))
    }

    fn tree(self, partition: &TransactionalPartitionHandle) -> Tree<'_, Self>
    where
        Self: Sized,
    {
        Tree {
            partition,
            mapping: self,
        }
    }
}

impl<'a, M> Tree<'a, M> {
    pub(super) fn len(
        &self,
        transaction: &ReadTransaction,
        prefix_key: Option<M::Key>,
    ) -> fjall::Result<usize>
    where
        M: Mapping,
    {
        Ok(self.prefix(transaction, prefix_key).count())
    }

    pub(super) fn insert(&self, transaction: &mut WriteTransaction, key: M::Key, value: M::Value)
    where
        M: Mapping,
    {
        transaction.insert(
            self.partition,
            self.mapping.key_path(key),
            value.to_bytes().into_owned(),
        );
    }

    pub(super) fn contains_key(
        &self,
        transaction: &mut WriteTransaction,
        key: M::Key,
    ) -> fjall::Result<bool>
    where
        M: Mapping,
    {
        transaction.contains_key(self.partition, self.mapping.key_path(key))
    }

    pub(super) fn read_contains_key(
        &self,
        transaction: &ReadTransaction,
        key: M::Key,
    ) -> fjall::Result<bool>
    where
        M: Mapping,
    {
        transaction.contains_key(self.partition, self.mapping.key_path(key))
    }

    pub(super) fn get(
        &self,
        transaction: &ReadTransaction,
        key: M::Key,
    ) -> Result<Option<M::Value>, M::Error>
    where
        M: Mapping,
    {
        transaction
            .get(self.partition, self.mapping.key_path(key))?
            .map(|slice| M::Value::parse(Cow::Owned(slice.to_vec())))
            .transpose()
            .map_err(M::Error::from)
    }

    pub(super) fn write_get(
        &self,
        transaction: &mut WriteTransaction,
        key: M::Key,
    ) -> Result<Option<M::Value>, M::Error>
    where
        M: Mapping,
    {
        transaction
            .get(self.partition, self.mapping.key_path(key))?
            .map(|slice| M::Value::parse(Cow::Owned(slice.to_vec())))
            .transpose()
            .map_err(M::Error::from)
    }

    pub(super) fn remove(&self, transaction: &mut WriteTransaction, key: M::Key)
    where
        M: Mapping,
    {
        transaction.remove(self.partition, self.mapping.key_path(key));
    }

    pub(super) fn prefix<'b>(
        &'b self,
        transaction: &'b ReadTransaction,
        key: Option<M::Key>,
    ) -> impl DoubleEndedIterator<Item = Result<(M::Key, M::Value), M::Error>> + 'static
    where
        M: Mapping,
    {
        transaction
            .prefix(self.partition, self.mapping.key_prefix(key))
            .filter_map(|res| {
                res.map_err(M::Error::from)
                    .and_then(|(key_slice, value_slice)| {
                        if let Ok(key) = M::parse_key_path(&key_slice) {
                            let value = M::Value::parse(Cow::Owned(value_slice.to_vec()))?;

                            Ok(Some((key, value)))
                        } else {
                            Ok(None)
                        }
                    })
                    .transpose()
            })
    }

    pub(super) fn range<'b>(
        &'b self,
        transaction: &'b ReadTransaction,
        (lower, upper): (Bound<M::Key>, Bound<M::Key>),
    ) -> impl DoubleEndedIterator<Item = Result<(M::Key, M::Value), M::Error>> + 'static
    where
        M: Mapping,
        M::Key: 'static,
    {
        let lower = lower.map(|value| self.mapping.key_path(value));
        let upper = upper.map(|value| self.mapping.key_path(value));

        transaction
            .range(self.partition, (lower, upper))
            .filter_map(|res| {
                res.map_err(M::Error::from)
                    .and_then(|(key_path_slice, value)| {
                        if let Ok(key) = M::parse_key_path(&key_path_slice) {
                            let value = M::Value::parse(Cow::Owned(value.to_vec()))?;

                            Ok(Some((key, value)))
                        } else {
                            Ok(None)
                        }
                    })
                    .transpose()
            })
    }
}
