#![cfg(system_deps)]

use common::{
    pict_rs_fjall_config, pict_rs_test_config, upload_form, with_pict_rs, Details, OkString,
    PictRsResult, UploadResponse,
};

mod common;

#[derive(Debug, serde::Deserialize)]
struct PageResponse {
    msg: OkString,
    page: Page,
}

#[derive(Debug, serde::Deserialize)]
struct Page {
    limit: u32,
    current: Option<String>,
    next: Option<String>,
    prev: Option<String>,
    hashes: Vec<Hash>,
}

#[derive(Debug, serde::Deserialize)]
struct Hash {
    hex: String,
    aliases: Vec<String>,
    details: Details,
}

#[test]
fn sled_can_page_uploaded_hashes() {
    let address = "127.0.0.1:9600";

    let mut config = pict_rs_test_config(address);

    config["server"]["max_file_count"] = 6.into();

    with_pict_rs(config, || can_page_upload_hashes(address)).unwrap();
}

#[test]
fn fjall_can_page_uploaded_hashes() {
    let address = "127.0.0.1:9601";

    let mut config = pict_rs_fjall_config(address);

    config["server"]["max_file_count"] = 6.into();

    with_pict_rs(config, || can_page_upload_hashes(address)).unwrap();
}

async fn can_page_upload_hashes(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form([
        "./client-examples/earth.gif",
        "./client-examples/cat.jpg",
        "./client-examples/scene.webp",
        "./client-examples/awoo.webp",
        "./client-examples/test.png",
        "./client-examples/earth.avif",
    ])
    .await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    assert!(response.status().is_success());

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    assert_eq!(upload.files.len(), 6);

    let mut slug = None;

    let mut page_count = 0;

    loop {
        let url = if let Some(slug) = slug {
            format!("http://{address}/internal/hashes?limit=2&slug={slug}")
        } else {
            format!("http://{address}/internal/hashes?limit=2")
        };

        let response = client
            .get(url)
            .header("X-Api-Token", "api-key")
            .send()
            .await
            .expect("send request");

        let body = response.bytes().await.expect("bytes");

        let PageResponse { page, .. } = match serde_json::from_slice::<PictRsResult<_>>(&body) {
            Ok(res) => res.unwrap(),
            Err(e) => {
                println!("{}", String::from_utf8_lossy(&body));
                panic!("{e} {e:?}");
            }
        };

        assert_eq!(page.limit, 2);
        assert_eq!(page.hashes.len(), 2);
        assert!(page.current.is_some());

        if !page.hashes.is_empty() {
            page_count += 1;
        }

        slug = page.next;

        if slug.is_none() {
            break;
        }
    }

    assert_eq!(page_count, 3);

    for file in &upload.files {
        let alias = &file.file;
        let delete_token = &file.delete_token;

        let response = client
            .delete(format!(
                "http://{address}/image/delete/{delete_token}/{alias}"
            ))
            .send()
            .await
            .expect("send request");

        assert!(response.status().is_success());
    }

    for file in &upload.files {
        let alias = &file.file;

        let response = client
            .get(format!("http://{address}/image/original/{alias}"))
            .send()
            .await
            .expect("download file");

        assert!(response.status().is_client_error());
    }

    tokio::time::sleep(std::time::Duration::from_millis(100)).await;

    slug = None;

    page_count = 0;

    loop {
        let url = if let Some(slug) = slug {
            format!("http://{address}/internal/hashes?limit=2&slug={slug}")
        } else {
            format!("http://{address}/internal/hashes?limit=2")
        };

        let response = client
            .get(url)
            .header("X-Api-Token", "api-key")
            .send()
            .await
            .expect("send request");

        let body = response.bytes().await.expect("bytes");

        let PageResponse { page, .. } = match serde_json::from_slice::<PictRsResult<_>>(&body) {
            Ok(res) => res.unwrap(),
            Err(e) => {
                println!("{}", String::from_utf8_lossy(&body));
                panic!("{e} {e:?}");
            }
        };

        assert_eq!(page.limit, 2);
        assert_eq!(page.hashes.len(), 0, "page should be empty\n{page:?}");
        assert!(page.current.is_none());

        if !page.hashes.is_empty() {
            page_count += 1;
        }

        slug = page.next;

        if slug.is_none() {
            break;
        }
    }

    assert_eq!(page_count, 0);
}
