#![cfg(system_deps)]

use common::{
    pict_rs_fjall_config, pict_rs_test_config, upload_form, with_pict_rs, PictRsResult,
    UploadResponse,
};

mod common;

#[test]
fn sled_can_upload_and_download_file() {
    let address = "127.0.0.1:8090";

    let config = pict_rs_test_config(address);

    with_pict_rs(config, || can_upload_and_download_file(address)).unwrap();
}

#[test]
fn fjall_can_upload_and_download_file() {
    let address = "127.0.0.1:8095";

    let config = pict_rs_fjall_config(address);

    with_pict_rs(config, || can_upload_and_download_file(address)).unwrap();
}

async fn can_upload_and_download_file(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form(["./client-examples/earth.gif"]).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    assert_eq!(upload.files.len(), 1);

    let alias = &upload.files[0].file;

    let response = client
        .get(format!("http://{address}/image/original/{alias}"))
        .send()
        .await
        .expect("download file");

    assert!(response.status().is_success(), "download failed");

    let length = response.bytes().await.expect("downlaod bytes").len();

    assert!(length > 0);
}

#[test]
fn sled_can_upload_and_download_multiple_files() {
    let address = "127.0.0.1:8091";

    let mut config = pict_rs_test_config(address);

    config["server"]["max_file_count"] = 4.into();

    with_pict_rs(config, || can_upload_and_download_multiple_files(address)).unwrap();
}

#[test]
fn fjall_can_upload_and_download_multiple_files() {
    let address = "127.0.0.1:8096";

    let mut config = pict_rs_fjall_config(address);

    config["server"]["max_file_count"] = 4.into();

    with_pict_rs(config, || can_upload_and_download_multiple_files(address)).unwrap();
}

async fn can_upload_and_download_multiple_files(address: &'static str) {
    let client = reqwest::Client::new();

    let files = [
        "./client-examples/earth.gif",
        "./client-examples/cat.jpg",
        "./client-examples/scene.webp",
        "./client-examples/test.png",
    ];

    let form = upload_form(files).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    assert_eq!(upload.files.len(), files.len());

    for file in upload.files {
        let alias = file.file;

        let response = client
            .get(format!("http://{address}/image/original/{alias}"))
            .send()
            .await
            .expect("download file");

        assert!(response.status().is_success(), "download failed");

        let length = response.bytes().await.expect("downlaod bytes").len();

        assert!(length > 0);
    }
}

#[test]
fn sled_can_delete_uploaded_file() {
    let address = "127.0.0.1:8092";

    let config = pict_rs_test_config(address);

    with_pict_rs(config, || can_delete_uploaded_file(address)).unwrap();
}

#[test]
fn fjall_can_delete_uploaded_file() {
    let address = "127.0.0.1:8097";

    let config = pict_rs_fjall_config(address);

    with_pict_rs(config, || can_delete_uploaded_file(address)).unwrap();
}

async fn can_delete_uploaded_file(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form(["./client-examples/earth.gif"]).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    assert_eq!(upload.files.len(), 1);

    let alias = &upload.files[0].file;

    let response = client
        .get(format!("http://{address}/image/original/{alias}"))
        .send()
        .await
        .expect("download file");

    assert!(response.status().is_success(), "download failed");

    let length = response.bytes().await.expect("downlaod bytes").len();

    assert!(length > 0);

    let delete_token = &upload.files[0].delete_token;

    let response = client
        .delete(format!(
            "http://{address}/image/delete/{delete_token}/{alias}"
        ))
        .send()
        .await
        .expect("send request");

    assert!(response.status().is_success());

    let response = client
        .get(format!("http://{address}/image/original/{alias}"))
        .send()
        .await
        .expect("download file");

    assert!(response.status().is_client_error());
}

#[test]
fn sled_cannot_upload_too_many_files() {
    let address = "127.0.0.1:8093";

    let mut config = pict_rs_test_config(address);

    config["server"]["max_file_count"] = 1.into();

    with_pict_rs(config, || cannot_upload_too_many_files(address)).unwrap();
}

#[test]
fn fjall_cannot_upload_too_many_files() {
    let address = "127.0.0.1:8098";

    let mut config = pict_rs_fjall_config(address);

    config["server"]["max_file_count"] = 1.into();

    with_pict_rs(config, || cannot_upload_too_many_files(address)).unwrap();
}

async fn cannot_upload_too_many_files(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form([
        "./client-examples/earth.gif",
        "./client-examples/cat.jpg",
        "./client-examples/scene.webp",
        "./client-examples/test.png",
    ])
    .await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response");

    match upload {
        PictRsResult::Ok(_) => panic!("request should have errored"),
        PictRsResult::Err { msg, code } => {
            assert_eq!(msg, "Too many files in request");
            assert_eq!(code, "file-upload-error");
        }
    }
}

#[test]
fn sled_cannot_upload_too_big_file() {
    let address = "127.0.0.1:8094";

    let mut config = pict_rs_test_config(address);

    config["media"]["max_file_size"] = 1.into();

    with_pict_rs(config, || cannot_upload_too_big_file(address)).unwrap();
}

#[test]
fn fjall_cannot_upload_too_big_file() {
    let address = "127.0.0.1:8099";

    let mut config = pict_rs_fjall_config(address);

    config["media"]["max_file_size"] = 1.into();

    with_pict_rs(config, || cannot_upload_too_big_file(address)).unwrap();
}

async fn cannot_upload_too_big_file(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form(["./client-examples/awoo.webp"]).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response");

    match upload {
        PictRsResult::Ok(_) => panic!("request should have errored"),
        PictRsResult::Err { msg, code } => {
            assert_eq!(msg, "File too large");
            assert_eq!(code, "validate-file-size");
        }
    }
}
