#![allow(dead_code)]

use std::{future::Future, str::FromStr};

use opentelemetry_sdk::propagation::TraceContextPropagator;
use tracing::subscriber::set_global_default;
use tracing_error::ErrorLayer;
use tracing_log::LogTracer;
use tracing_subscriber::{fmt::format::FmtSpan, layer::SubscriberExt, EnvFilter, Layer, Registry};

pub fn with_pict_rs<F, Fut>(config: serde_json::Value, callback: F) -> color_eyre::Result<()>
where
    F: Fn() -> Fut,
    Fut: Future<Output = ()>,
{
    tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()?
        .block_on(async move {
            tokio::task::LocalSet::new()
                .run_until(async move {
                    if init_tracing().is_err() {
                        tracing::info!("tracing already enabled");
                    };

                    let mut pict_rs_handle = spawn_pict_rs(config);

                    // give time to spin up
                    tokio::time::sleep(std::time::Duration::from_millis(500)).await;

                    tokio::select! {
                        _ = (callback)() => {
                            pict_rs_handle.abort();
                            let _ = pict_rs_handle.await;
                        }
                        res = &mut pict_rs_handle => {
                            res.unwrap();
                        }
                    }

                    Ok(())
                })
                .await
        })
}

pub async fn upload_form<I>(files: I) -> reqwest::multipart::Form
where
    I: IntoIterator<Item = &'static str>,
{
    let mut form = reqwest::multipart::Form::new();

    for file in files {
        form = form.file("images[]", file).await.expect("read file");
    }

    form
}

pub fn pict_rs_test_config(address: &str) -> serde_json::Value {
    let directory = format!("/tmp/pict-rs-test/{}", uuid::Uuid::now_v7());

    serde_json::json!({
         "server": {
             "address": address,
             "temporary_directory": format!("{directory}/tmp"),
             "api_key": "api-key"
         },
         "repo": {
             "type": "sled",
             "path": format!("{directory}/sled-repo")
         },
         "store": {
             "type": "filesystem",
             "path": format!("{directory}/files")
         }

    })
}

pub fn pict_rs_fjall_config(address: &str) -> serde_json::Value {
    let directory = format!("/tmp/pict-rs-test/{}", uuid::Uuid::now_v7());

    serde_json::json!({
         "server": {
             "address": address,
             "temporary_directory": format!("{directory}/tmp"),
             "api_key": "api-key"
         },
         "repo": {
             "type": "fjall",
             "fjall_path": format!("{directory}/fjall-repo")
         },
         "store": {
             "type": "filesystem",
             "path": format!("{directory}/files")
         }

    })
}

fn spawn_pict_rs(config: serde_json::Value) -> tokio::task::JoinHandle<()> {
    tokio::task::spawn_local(async move {
        pict_rs::ConfigSource::memory(config)
            .init::<String>(None)
            .expect("init pict-rs config")
            .install_crypto_provider()
            .run()
            .await
            .expect("run pict-rs")
    })
}

fn init_tracing() -> color_eyre::Result<()> {
    LogTracer::init()?;

    opentelemetry::global::set_text_map_propagator(TraceContextPropagator::new());

    let format_layer = tracing_subscriber::fmt::layer()
        .with_test_writer()
        .with_span_events(FmtSpan::NEW | FmtSpan::CLOSE)
        .with_filter(EnvFilter::from_str("info")?);

    let subscriber = Registry::default()
        .with(format_layer)
        .with(ErrorLayer::default());

    set_global_default(subscriber)?;

    Ok(())
}

#[derive(Debug, serde::Deserialize)]
#[serde(untagged)]
pub enum PictRsResult<T> {
    Ok(T),
    Err { msg: String, code: String },
}

impl<T> PictRsResult<T> {
    pub fn unwrap(self) -> T {
        match self {
            Self::Ok(t) => t,
            Self::Err { msg, code } => panic!("{code}: {msg}"),
        }
    }
}

#[derive(Debug, serde::Deserialize)]
pub enum OkString {
    #[serde(rename = "ok")]
    Ok,
}

#[derive(Debug, serde::Deserialize)]
pub struct UploadResponse {
    pub msg: OkString,

    pub files: Vec<File>,
}

#[derive(Debug, serde::Deserialize)]
pub struct File {
    pub delete_token: String,
    pub file: String,
    pub details: Details,
}

#[derive(Debug, serde::Deserialize)]
pub struct Details {
    pub width: u16,
    pub height: u16,
    pub frames: Option<u32>,
    pub blurhash: String,
    pub content_type: String,
    #[serde(with = "time::serde::rfc3339")]
    pub created_at: time::OffsetDateTime,
}
