#![cfg(system_deps)]

use std::collections::HashMap;

use common::{
    pict_rs_fjall_config, pict_rs_test_config, upload_form, with_pict_rs, PictRsResult,
    UploadResponse,
};

mod common;

#[test]
fn sled_reads_image_details_from_uploads() {
    let address = "127.0.0.1:9300";

    let mut config = pict_rs_test_config(address);

    config["server"]["max_file_count"] = 10.into();

    with_pict_rs(config, || reads_image_details_from_uploads(address)).unwrap();
}

#[test]
fn fjall_reads_image_details_from_uploads() {
    let address = "127.0.0.1:9301";

    let mut config = pict_rs_fjall_config(address);

    config["server"]["max_file_count"] = 10.into();

    with_pict_rs(config, || reads_image_details_from_uploads(address)).unwrap();
}

async fn reads_image_details_from_uploads(address: &'static str) {
    struct Case {
        file: &'static str,
        width: u16,
        height: u16,
        frames: Option<u32>,
        blurhash: &'static str,
        content_type: &'static str,
    }

    impl Case {
        fn check(&self, details: &common::Details) {
            assert_eq!(self.width, details.width, "invalid width for {}", self.file);
            assert_eq!(
                self.height, details.height,
                "invalid height for {}",
                self.file
            );
            assert_eq!(
                self.frames, details.frames,
                "invalid frames for {}",
                self.file
            );
            assert_eq!(
                self.blurhash, details.blurhash,
                "invalid blurhash for {}",
                self.file
            );
            assert_eq!(
                self.content_type, details.content_type,
                "invalid content type for {}",
                self.file
            );
        }
    }

    let cases = [
        Case {
            file: "./client-examples/earth.gif",
            width: 400,
            height: 400,
            frames: Some(44),
            blurhash: "KD7n8nWF0M%1a#M|0MoI^%",
            content_type: "image/gif",
        },
        Case {
            file: "./client-examples/awoo.webp",
            width: 3000,
            height: 4000,
            frames: None,
            blurhash: "TCG8.f~957XNV@?ZORI]RP-.ozIU",
            content_type: "image/webp",
        },
        Case {
            file: "./client-examples/earth.avif",
            width: 600,
            height: 600,
            frames: Some(44),
            blurhash: "KC7U3RWF0M%1a#Ip0MoI^%",
            content_type: "image/avif",
        },
        Case {
            file: "./client-examples/cat.jpg",
            width: 1200,
            height: 1387,
            frames: None,
            blurhash: "KmKn9x4T.8D%j[V@o~RkjY",
            content_type: "image/jpeg",
        },
        Case {
            file: "./client-examples/scene.webp",
            width: 550,
            height: 368,
            frames: None,
            blurhash: "L_C@s#R-RPae.TRkWCof.8WFR.ay",
            content_type: "image/webp",
        },
        Case {
            file: "./client-examples/test.png",
            width: 640,
            height: 360,
            frames: None,
            blurhash: "MPB;OHN6XdaJeswBNOs,W9R*E7t7nPogo#",
            content_type: "image/png",
        },
    ];

    let client = reqwest::Client::new();

    let form = upload_form(cases.iter().map(|c| c.file)).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    let cases = cases
        .iter()
        .map(|c| (c.blurhash, c))
        .collect::<HashMap<_, _>>();

    for file in upload.files.iter() {
        let case = cases
            .get(file.details.blurhash.as_str())
            .expect("blurhash in cases");

        case.check(&file.details);
    }
}
