#![cfg(system_deps)]

use common::{pict_rs_test_config, upload_form, with_pict_rs, PictRsResult, UploadResponse};

mod common;

#[test]
fn cannot_upload_too_wide_image() {
    let address = "127.0.0.1:9000";

    let mut config = pict_rs_test_config(address);

    config["media"]["image"]["max_width"] = 100.into();

    with_pict_rs(config, || async {
        let client = reqwest::Client::new();

        let form = upload_form(["./client-examples/cat.jpg"]).await;

        let response = client
            .post(format!("http://{address}/image"))
            .multipart(form)
            .send()
            .await
            .expect("send request");

        let upload = response
            .json::<PictRsResult<UploadResponse>>()
            .await
            .expect("valid response");

        match upload {
            PictRsResult::Ok(_) => panic!("request should have errored"),
            PictRsResult::Err { msg, code } => {
                assert_eq!(msg, "Too wide");
                assert_eq!(code, "validate-width");
            }
        }
    })
    .unwrap();
}

#[test]
fn cannot_upload_too_tall_image() {
    let address = "127.0.0.1:9001";

    let mut config = pict_rs_test_config(address);

    config["media"]["image"]["max_height"] = 100.into();

    with_pict_rs(config, || async {
        let client = reqwest::Client::new();

        let form = upload_form(["./client-examples/cat.jpg"]).await;

        let response = client
            .post(format!("http://{address}/image"))
            .multipart(form)
            .send()
            .await
            .expect("send request");

        let upload = response
            .json::<PictRsResult<UploadResponse>>()
            .await
            .expect("valid response");

        match upload {
            PictRsResult::Ok(_) => panic!("request should have errored"),
            PictRsResult::Err { msg, code } => {
                assert_eq!(msg, "Too tall");
                assert_eq!(code, "validate-height");
            }
        }
    })
    .unwrap();
}

#[test]
fn cannot_upload_too_much_area_image() {
    let address = "127.0.0.1:9002";

    let mut config = pict_rs_test_config(address);

    config["media"]["image"]["max_area"] = 100.into();

    with_pict_rs(config, || async {
        let client = reqwest::Client::new();

        let form = upload_form(["./client-examples/cat.jpg"]).await;

        let response = client
            .post(format!("http://{address}/image"))
            .multipart(form)
            .send()
            .await
            .expect("send request");

        let upload = response
            .json::<PictRsResult<UploadResponse>>()
            .await
            .expect("valid response");

        match upload {
            PictRsResult::Ok(_) => panic!("request should have errored"),
            PictRsResult::Err { msg, code } => {
                assert_eq!(msg, "Too many pixels");
                assert_eq!(code, "validate-area");
            }
        }
    })
    .unwrap();
}

#[test]
fn cannot_upload_too_large_image() {
    let address = "127.0.0.1:9003";

    let mut config = pict_rs_test_config(address);

    config["media"]["image"]["max_file_size"] = 1.into();

    with_pict_rs(config, || async {
        let client = reqwest::Client::new();

        let form = upload_form(["./client-examples/awoo.webp"]).await;

        let response = client
            .post(format!("http://{address}/image"))
            .multipart(form)
            .send()
            .await
            .expect("send request");

        let upload = response
            .json::<PictRsResult<UploadResponse>>()
            .await
            .expect("valid response");

        match upload {
            PictRsResult::Ok(_) => panic!("request should have errored"),
            PictRsResult::Err { msg, code } => {
                assert_eq!(msg, "Filesize too large");
                assert_eq!(code, "validate-file-size");
            }
        }
    })
    .unwrap();
}
