#![cfg(system_deps)]

use common::{
    pict_rs_fjall_config, pict_rs_test_config, upload_form, with_pict_rs, Details, PictRsResult,
    UploadResponse,
};

mod common;

#[test]
fn sled_can_identity_process_image() {
    let address = "127.0.0.1:9400";

    let config = pict_rs_test_config(address);

    with_pict_rs(config, || can_identity_process_image(address)).unwrap();
}

#[test]
fn fjall_can_identity_process_image() {
    let address = "127.0.0.1:9405";

    let config = pict_rs_fjall_config(address);

    with_pict_rs(config, || can_identity_process_image(address)).unwrap();
}

async fn can_identity_process_image(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form(["./client-examples/cat.jpg"]).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    let alias = &upload.files[0].file;
    let original_details = &upload.files[0].details;

    let response = client
        .get(format!(
            "http://{address}/image/process.jpg?alias={alias}&identity=true"
        ))
        .send()
        .await
        .expect("send request");

    assert!(response.status().is_success());

    let bytes = response.bytes().await.expect("response body");

    assert!(!bytes.is_empty());

    let details_response = client
        .get(format!(
            "http://{address}/image/details/process.jpg?alias={alias}&identity=true"
        ))
        .send()
        .await
        .expect("send request");

    assert!(details_response.status().is_success());

    let process_details = details_response
        .json::<Details>()
        .await
        .expect("valid response");

    assert_eq!(original_details.width, process_details.width);
    assert_eq!(original_details.height, process_details.height);
    assert_eq!(original_details.frames, process_details.frames);
}

#[test]
fn sled_can_crop_process_image() {
    let address = "127.0.0.1:9401";

    let config = pict_rs_test_config(address);

    with_pict_rs(config, || can_crop_process_image(address)).unwrap();
}

#[test]
fn fjall_can_crop_process_image() {
    let address = "127.0.0.1:9406";

    let config = pict_rs_fjall_config(address);

    with_pict_rs(config, || can_crop_process_image(address)).unwrap();
}

async fn can_crop_process_image(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form(["./client-examples/cat.jpg"]).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    let alias = &upload.files[0].file;
    let original_details = &upload.files[0].details;

    let response = client
        .get(format!(
            "http://{address}/image/process.jpg?alias={alias}&crop=1x1"
        ))
        .send()
        .await
        .expect("send request");

    assert!(response.status().is_success());

    let bytes = response.bytes().await.expect("response body");

    assert!(!bytes.is_empty());

    let details_response = client
        .get(format!(
            "http://{address}/image/details/process.jpg?alias={alias}&crop=1x1"
        ))
        .send()
        .await
        .expect("send request");

    assert!(details_response.status().is_success());

    let process_details = details_response
        .json::<Details>()
        .await
        .expect("valid response");

    assert_ne!(original_details.width, original_details.height);
    assert_eq!(process_details.width, process_details.height);
}

#[test]
fn sled_can_resize_process_image() {
    let address = "127.0.0.1:9402";

    let config = pict_rs_test_config(address);

    with_pict_rs(config, || can_resize_process_image(address)).unwrap();
}

#[test]
fn fjall_can_resize_process_image() {
    let address = "127.0.0.1:9407";

    let config = pict_rs_fjall_config(address);

    with_pict_rs(config, || can_resize_process_image(address)).unwrap();
}

async fn can_resize_process_image(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form(["./client-examples/cat.jpg"]).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    let alias = &upload.files[0].file;
    let original_details = &upload.files[0].details;

    let response = client
        .get(format!(
            "http://{address}/image/process.jpg?alias={alias}&resize=512"
        ))
        .send()
        .await
        .expect("send request");

    assert!(response.status().is_success());

    let bytes = response.bytes().await.expect("response body");

    assert!(!bytes.is_empty());

    let details_response = client
        .get(format!(
            "http://{address}/image/details/process.jpg?alias={alias}&resize=512"
        ))
        .send()
        .await
        .expect("send request");

    assert!(details_response.status().is_success());

    let process_details = details_response
        .json::<Details>()
        .await
        .expect("valid response");

    assert!(original_details.width > 512);
    assert!(original_details.height > 512);
    assert!(process_details.width <= 512);
    assert!(process_details.height <= 512);

    // check aspect ratios
    assert!(
        (original_details.height as f64 / original_details.width as f64)
            - (process_details.height as f64 / process_details.width as f64)
            <= 0.001
    )
}

#[test]
fn sled_can_resize_area_process_image() {
    let address = "127.0.0.1:9403";

    let config = pict_rs_test_config(address);

    with_pict_rs(config, || can_resize_area_process_image(address)).unwrap();
}

#[test]
fn fjall_can_resize_area_process_image() {
    let address = "127.0.0.1:9408";

    let config = pict_rs_fjall_config(address);

    with_pict_rs(config, || can_resize_area_process_image(address)).unwrap();
}

async fn can_resize_area_process_image(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form(["./client-examples/cat.jpg"]).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    let alias = &upload.files[0].file;
    let original_details = &upload.files[0].details;

    let response = client
        .get(format!(
            "http://{address}/image/process.jpg?alias={alias}&resize=.a625"
        ))
        .send()
        .await
        .expect("send request");

    assert!(response.status().is_success());

    let bytes = response.bytes().await.expect("response body");

    assert!(!bytes.is_empty());

    let details_response = client
        .get(format!(
            "http://{address}/image/details/process.jpg?alias={alias}&resize=.a625"
        ))
        .send()
        .await
        .expect("send request");

    assert!(details_response.status().is_success());

    let process_details = details_response
        .json::<Details>()
        .await
        .expect("valid response");

    assert!(original_details.width > 40);
    assert!(original_details.height > 40);
    assert!(process_details.width <= 40);
    assert!(process_details.height <= 40);

    // check aspect ratios
    assert!(
        (original_details.height as f64 / original_details.width as f64)
            - (process_details.height as f64 / process_details.width as f64)
            <= 0.001
    )
}

#[test]
fn sled_can_thumbnail_process_image() {
    let address = "127.0.0.1:9404";

    let config = pict_rs_test_config(address);

    with_pict_rs(config, || can_thumbnail_process_image(address)).unwrap();
}

#[test]
fn fjall_can_thumbnail_process_image() {
    let address = "127.0.0.1:9409";

    let config = pict_rs_fjall_config(address);

    with_pict_rs(config, || can_thumbnail_process_image(address)).unwrap();
}

async fn can_thumbnail_process_image(address: &'static str) {
    let client = reqwest::Client::new();

    let form = upload_form(["./client-examples/cat.jpg"]).await;

    let response = client
        .post(format!("http://{address}/image"))
        .multipart(form)
        .send()
        .await
        .expect("send request");

    let upload = response
        .json::<PictRsResult<UploadResponse>>()
        .await
        .expect("valid response")
        .unwrap();

    let alias = &upload.files[0].file;
    let original_details = &upload.files[0].details;

    let response = client
        .get(format!(
            "http://{address}/image/process.jpg?alias={alias}&thumbnail=512"
        ))
        .send()
        .await
        .expect("send request");

    assert!(response.status().is_success());

    let bytes = response.bytes().await.expect("response body");

    assert!(!bytes.is_empty());

    let details_response = client
        .get(format!(
            "http://{address}/image/details/process.jpg?alias={alias}&thumbnail=512"
        ))
        .send()
        .await
        .expect("send request");

    assert!(details_response.status().is_success());

    let process_details = details_response
        .json::<Details>()
        .await
        .expect("valid response");

    assert!(original_details.width > 512);
    assert!(original_details.height > 512);
    assert!(process_details.width <= 512);
    assert!(process_details.height <= 512);

    // check aspect ratios
    assert!(
        (original_details.height as f64 / original_details.width as f64)
            - (process_details.height as f64 / process_details.width as f64)
            <= 0.001
    )
}
